# put the just-built tkwant into the path
import sys
from distutils.util import get_platform
sys.path.insert(0, "../../../build/lib.{0}-{1}.{2}"
                .format(get_platform(), *sys.version_info[:2]))



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2020 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.
"""Classes for solving the time-dependent Schrödinger equation."""

import numpy as np
import scipy.integrate

from . cimport kernels
from . import kernels


__all__ = ['Scipy']


class Scipy:
    """Solve the time-dependent Schrödinger equation using `scipy.integrate`.

    This solver will currently only work with the 'dopri5' and 'dop853'
    integrators, as these are the only re-entrant ones.

    Parameters
    ----------
    kernel : `tkwant.onebody.kernels.Kernel`
    integrator : `scipy.integrate._ode.IntegratorBase`, default: dopri5
        The integrator to use with this solver.
    **integrator_options
        Options to pass when instantiating the integrator.

    See Also
    --------
    scipy.integrate.ode
    """

    _default_options = {'atol': 1E-9, 'rtol': 1E-9, 'nsteps': int(1E9)}

    def __init__(self, kernel, integrator=scipy.integrate._ode.dopri5, **integrator_options):
        self.kernel = kernel
        # allocate storage for kernel output
        self._rhs_out = np.empty((kernel.size,), dtype=complex)

        options = dict(self._default_options)
        options.update(integrator_options)
        self._integrator = integrator(**options)
        # Factor 2 because Scipy integrators expect real arrays
        self._integrator.reset(2 * self.kernel.size, has_jac=False)

    def _rhs(self, t, y):
        # Kernel expects complex, Scipy expects real
        self.kernel.rhs(y.view(complex), self._rhs_out, t)
        return self._rhs_out.view(float)

    def __call__(self, psi, time, next_time):
        if time == next_time:
            return psi
        # psi is complex, Scipy expects real
        next_psi, final_time = self._integrator.run(
            self._rhs, lambda: None, psi.view(float), time, next_time, (), ())
        if not self._integrator.success:
            raise RuntimeError('Integration failed between {} and {}'
                               .format(time, next_time))
        assert final_time == next_time
        return next_psi.view(complex)


default = Scipy



# -*- coding: utf-8 -*-
# Copyright 2016-2019 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.
"""Tools for dealing with time-dependent infinite, periodic leads.

Also contains tools for dealing with boundary conditions that simulate
the effect of leads when solving the time-dependent Schrödinger equation.
"""

import abc
import bisect
import functools as ft
import collections
import inspect
import cmath
from math import ceil
import numpy as np
from scipy import sparse as sp
from scipy.optimize import brentq

import kwant
import kwantspectrum
import tinyarray as ta
from . import _common, _logging


__all__ = ['add_voltage', 'SimpleBoundary', 'MonomialAbsorbingBoundary',
           'GenericAbsorbingBoundary', 'automatic_boundary',
           'AbsorbingReflectionSolver', 'AnalyzeReflection',
           'AnalyzeReflectionMonomial']


# set module logger
logger = _logging.make_logger(name=__name__)
log_func = _logging.log_func(logger)


# TODO: change this when Kwant can have further than nearest-neighbor
#       hoppings in leads
@log_func
def add_voltage(syst, lead, phase):
    r"""Add a time-dependent voltage to a lead.

    Add a lead unit cell to the ``syst`` where ``lead`` is
    attached, and couple this to ``syst`` with a time-dependent
    hopping that effectively adds a voltage to the lead.

    Parameters
    ----------
    syst : `kwant.builder.Builder`
        The central system, with its leads already attached.
        Modified on return.
    lead : int
        The number of the lead on which to apply the voltage.
    phase : callable
        Function specifying the anti-derivative of the voltage.
        Takes takes the same extra arguments as the
        Hamiltonian value functions, starting with the time.
        Returns either a scalar or a one-dimensional sequence:
        one element per orbital on a site. Must return 0
        (or a sequence of zeros) at time 0.

    Returns
    -------
    tuple
        The sites added to the system.

    Raises
    ------
    ValueError
        If the ``phase`` argument is not callable, or if sites in the
        lead interface belong to different domains

    Notes
    -----
    Formally, this function adds a single lead cell to the
    system and attaches the previous lead interface to it with
    a hopping :math:`(1 \otimes \exp[iX(t)])V`, where V is the hopping
    between lead unit cells, 1 is the unit matrix over the interface sites
    and X(t) is a square diagonal matrix over the orbitals on
    one site, specified by the ``phase`` parameter.  This
    corresponds to adding a voltage X'(t) to each site in the
    lead. In the most common case of 1 orbital per site X(t) is
    a scalar.

    This function only works if the lead was attached with ``attach_lead``,

    If any exceptions are raised in this function there are no
    guarantees that the system was not modified.
    """
    if not isinstance(syst, kwant.builder.Builder):
        raise TypeError('"syst" must be an instance of ``kwant.builder.Builder``'
                        '(not a finalized system)')

    lead_builder = syst.leads[lead].builder
    interface = syst.leads[lead].interface
    sym = lead_builder.symmetry
    interface_dom = sym.which(interface[0])
    # check for valid inputs
    if not callable(phase):
        raise ValueError('Phase function is not callable')
    # sanity check that interface sites are all in the same domain,
    # in any case, if this is not true Kwant will not be able to
    # finalise the system
    if not all(sym.which(s) == interface_dom for s in interface):
        raise ValueError('Some interface sites belong to different domains')
    # check that there are not hoppings > nearest neighbor cells
    for hopping in lead_builder.hoppings():
        if not -1 <= sym.which(hopping[1])[0] <= 1:
            msg = ('The following hopping connects non-neighboring lead '
                   'unit cells. Only nearest-cell hoppings are allowed '
                   '(consider increasing the lead period).\n{0}')
            raise ValueError(msg.format(hopping))

    # phase *from* the system *to* the lead (new cell)
    # called with a hopping and whether this is *from* the system *to*
    # the lead (new cell) or vice versa, returns a hopping value function

    # map sites to domain of the new cell
    def move(x):
        return sym.act(interface_dom + 1, sym.act(-sym.which(x), x))

    # copy over sites
    for site in lead_builder.H:
        syst[move(site)] = lead_builder[site]

    # copy over hoppings
    for (to_site, from_site), value in lead_builder.hopping_value_pairs():
        if sym.which(to_site) == sym.which(from_site):  # intra-cell hopping
            syst[move(to_site), move(from_site)] = value
        else:  # inter-cell hopping
            system_to_lead = sym.which(to_site)[0] > sym.which(from_site)[0]
            if system_to_lead:
                from_site = sym.act(interface_dom - sym.which(from_site), from_site)
                to_site = move(to_site)
            else:
                to_site = sym.act(interface_dom - sym.which(to_site), to_site)
                from_site = move(from_site)
            syst[to_site, from_site] = _make_time_dependent_hopping(phase, value, system_to_lead)

    syst.leads[lead].interface = tuple(map(move, interface))
    return tuple(map(move, lead_builder.H))


# ---------- Boundary Conditions

def _return_true(x):
    return True


class _TimeIsValid:
    # Do not allow us to evolve beyond tmax
    def __init__(self, tmax):
        self.tmax = tmax

    def __call__(self, time):
        return time < self.tmax


class EvaluatedBoundary(collections.namedtuple('_System', ['hamiltonian',
                                                           'to_slices',
                                                           'from_slices',
                                                           'solution_is_valid',
                                                           'time_is_valid'])):
    """Boundary conditions evaluated for a lead.

    Parameters
    ----------
    hamiltonian : `~scipy.sparse.coo_matrix`
        The Hamiltonian matrix of the boundary conditions evaluated over
        the lead.
    to_slices, from_slices : sequence of `tinyarray.array`
        Each array contains a pair of integers: a slice into ``hamiltonian``
        which we wish to connect *to* (or *from*) the Hamiltonian of the
        central region.
    solution_is_valid : callable
        Function that takes a 1D array of the same size as ``hamiltonian``
        (i.e the solution inside the boundary conditions), and returns
        True if there are no spurious reflections detected.
    time_is_valid : sequence of callable
        Function that takes a float (time) and returns True if the
        boundary conditions will remain valid up till this time.

    Notes
    -----
    ``to_slices`` and ``from_slices`` are used when constructing the
    Hamiltonian for the central system + boundary conditions. Usually The first
    slices will be over the first cell of the lead. Subsequent entries in
    ``from_slices`` can be used, for example, when constructing two "copies" of
    a boundary where the second copy should be affected by what happens in the
    central region, but should not itself have any back-action on the central
    region (see section C of [1] for details).

    [1]: http://arxiv.org/abs/1510.05967
    """


class BoundaryBase(metaclass=abc.ABCMeta):
    """ABC boundary conditions for the time-dependent Schrödinger equation."""

    @abc.abstractmethod
    def __call__(self, lead, params=None):
        """Generate boundary conditions for a single lead.

        These boundary condition matrices must be formatted such that
        they can be coupled to the central system Hamiltonian via
        ``lead.inter_cell_hopping`` in the first ``lead.n_cells``
        orbitals of the matrix corresponding to ``lead``.

        Parameters
        ----------
        lead : `~kwant.system.InfiniteSystem`
            The lead for which to generate boundary conditions.
        params : dict, optional
            Extra parameters for Hamiltonian value functions.

        Returns
        -------
        `_EvaluatedBoundary`
        """
        pass


class SimpleBoundary(BoundaryBase):
    """Boundary conditions consisting of N lead unit cells.

    Parameters
    ----------
    num_cells : int, optional
        The number of lead unit cells to add. Mutually exclusive with
        ``tmax``.
    tmax : float, optional
        The maximum time up to which we wish to simulate with these
        boundary conditions. Mutually exclusive with ``num_cells``
    """

    def __init__(self, num_cells=None, tmax=None):
        self.num_total_cells = self.tmax = None

        if num_cells is not None:
            if num_cells <= 0:
                raise ValueError('Number of unit cells must be positive.')
            self.num_total_cells = int(num_cells)

        if tmax is not None:
            if tmax <= 0:
                raise ValueError('Maximum time must be positive.')
            self.tmax = float(tmax + 1)  # 1 to prevent rounding problems

        if self.num_total_cells is None and self.tmax is None:
            raise TypeError('Either `num_cells` or `tmax` must be provided.')
        elif self.num_total_cells is not None and self.tmax is not None:
            raise TypeError('`num_cells` and `tmax` are mutually exclusive.')

    def __call__(self, lead, params=None):
        H_cell = lead.cell_hamiltonian(params=params, sparse=True)
        V_cell = lead.inter_cell_hopping(params=params, sparse=True)
        V_cell = _make_square_matrix(V_cell)
        cell_norbs = H_cell.shape[0]

        # twice the spectral norm of V_cell gives max velocity
        max_velocity = (2 * np.linalg.norm(V_cell.todense(), ord=2))

        tmax = self.tmax or self.num_total_cells / max_velocity
        # add 1 to prevent off-by-one error due to inter-cell hoppings
        num_cells = (self.num_total_cells
                     or int(np.ceil(self.tmax * max_velocity)) + 1)

        assert tmax is not None

        H = _make_block_tridiagonal(H_cell, V_cell,
                                    V_cell.conjugate().transpose(),
                                    num_cells)

        # With these boundary conditions there is no way to explicitly
        # check if the solution is good for a given lead_psi.
        solution_is_valid = _return_true
        time_is_valid = _TimeIsValid(tmax)

        return EvaluatedBoundary(hamiltonian=H,
                                 to_slices=[ta.array([0, cell_norbs])],
                                 from_slices=[ta.array([0, cell_norbs])],
                                 solution_is_valid=solution_is_valid,
                                 time_is_valid=time_is_valid)


class AbsorbingBoundary(SimpleBoundary):

    def __call__(self, lead, params=None):
        # get the regular boundary conditions by calling superclass
        simple_boundary = super().__call__(lead, params)

        # make diagonal matrix with entries increasing as we go to further into
        # the lead, but with entries constant over a given cell.
        bc_diags = [self._absorb(cell) for cell in range(self.num_cells)]
        bc_diags = [0] * self.num_buffer_cells + bc_diags  # add buffer cells
        cell_norbs = _get_slice(lead, lead.cell_size).start
        S = sp.kron(sp.diags(bc_diags, offsets=0), sp.eye(cell_norbs))
        H = simple_boundary.hamiltonian - 1j * S

        # With these boundary conditions there is no way to explicitly
        # check if the solution is good for a given lead_psi.
        solution_is_valid = _return_true

        # apriori there is no "max time" with absorbing boundary conditions
        time_is_valid = _return_true

        return EvaluatedBoundary(hamiltonian=H,
                                 to_slices=simple_boundary.to_slices,
                                 from_slices=simple_boundary.from_slices,
                                 solution_is_valid=solution_is_valid,
                                 time_is_valid=time_is_valid)

    @abc.abstractmethod
    def _absorb(self, cell):
        """Return the absorbing potential in the given cell."""
        pass


class MonomialAbsorbingBoundary(AbsorbingBoundary):
    """Absorbing boundary conditions consisting of N lead unit cells.

    The absorbing region has an imaginary potential applied to it.
    The magnitude of the imaginary potential grows according to
    ``n**degree`` where ``n`` is the index (starting from 0) of the
    lead cell counting from the central region.

    Parameters
    ----------
    num_cells : int
        The number of lead unit cells over which the absorbing potential
        increases.
    strength : float
        The strength of the boundary conditions. Formally
        this is the area underneath the monomial curve.
    degree : int
        The degree of the absorbing monomial.
    num_buffer_cells : int, optional
        If provided, adds this many lead cells to the start
        of the boundary conditions with no absorbing potential
        applied.
    """

    def __init__(self, num_cells, strength, degree, num_buffer_cells=0):
        if num_cells < 0:
            raise ValueError('Cannot add a negative number of absorbing cells')
        if num_buffer_cells < 0:
            raise ValueError('Cannot add a negative number of buffer cells')
        if degree < 0:
            raise ValueError('Absorbing boundary conditions cannot be given '
                             'by a monomial of negative degree.')
        super().__init__(num_cells + num_buffer_cells)
        self.num_cells = int(num_cells)
        self.strength = strength
        self.degree = int(degree)
        self.num_buffer_cells = int(num_buffer_cells)

    def _absorb(self, cell):
        n = self.degree
        return (n + 1) * self.strength * (cell**n / self.num_cells**(n + 1))


class GenericAbsorbingBoundary(AbsorbingBoundary):
    """Make an absorbing boundary from a user provided-absorbing potential.

    Parameters
    ----------
    imaginary_potential : callable
        User-provided absorbing potential.
    num_cells : int
        The number of lead unit cells over which the absorbing potential
        increases.
    num_buffer_cells : int, optional
        If provided, adds this many lead cells to the start
        of the boundary conditions with no absorbing potential applied.
    """

    def __init__(self, num_cells, imaginary_potential, num_buffer_cells=0):

        if not callable(imaginary_potential):
            raise ValueError('Absorbing potential is not callable')
        if num_cells < 0:
            raise ValueError('Cannot add a negative number of absorbing cells')
        if num_buffer_cells < 0:
            raise ValueError('Cannot add a negative number of buffer cells')

        super().__init__(num_cells + num_buffer_cells)
        self.num_cells = int(num_cells)
        self.num_buffer_cells = int(num_buffer_cells)
        self._imaginary_potential = imaginary_potential

    def _absorb(self, cell):
        return self._imaginary_potential(cell / self.num_cells) / self.num_cells


# TODO: add boundary conditions with a second "copy" that we can use to tell
#       if our solution has any spurious reflection.


# ---------- Internal functions
# TODO: move these to Cython

def _set_signature(func, params):
    """Set the signature of 'func'.

    Parameters
    ----------
    func : callable
    params: sequence of str
        Parameter names to put in the signature. These will be added as
        'POSITIONAL_ONLY' type parameters.
    """
    params = [inspect.Parameter(name, inspect.Parameter.POSITIONAL_ONLY)
              for name in params]
    func.__signature__ = inspect.Signature(params)


class _PhaseFactor:
    def __init__(self, phase):
        self.phase = phase

    def __call__(self, args):
        try:
            val = np.asarray(self.phase(*args))
            # if a scalar, `diag` will not work
            if len(val.shape) == 0:
                return cmath.exp(1j * val)
            return np.diag(np.exp(1j * val))
        except Exception as exc:
            msg = 'Error while evaluating user-supplied '\
                  'phase function "{}": '.format(self.phase.__name__)
            if exc.args:
                msg += str(exc.args[0])
            raise RuntimeError(msg) from exc


class _CallableHoppingSystemToLead:

    def __init__(self, phase_factor, old_hopping, len_hopp, mapping):

        self.phase_factor = phase_factor
        self.old_hopping = old_hopping
        self.len_hopp = len_hopp
        self.mapping = mapping

    def __call__(self, i, j, *args):
        old_hopping_args = args[:self.len_hopp]
        phase_args = tuple(map(lambda i: args[i], self.mapping))
        return np.dot(self.phase_factor(phase_args),
                      self.old_hopping(i, j, *old_hopping_args))


class _CallableHoppingLeadToSystem:

    def __init__(self, phase_factor, old_hopping, len_hopp, mapping):

        self.phase_factor = phase_factor
        self.old_hopping = old_hopping
        self.len_hopp = len_hopp
        self.mapping = mapping

    def __call__(self, i, j, *args):
        old_hopping_args = args[:self.len_hopp]
        phase_args = tuple(map(lambda i: args[i], self.mapping))
        return np.dot(self.old_hopping(i, j, *old_hopping_args),
                      self.phase_factor(phase_args).conjugate())


class _StaticHoppingSystemToLead:

    def __init__(self, phase_factor, old_hopping):

        self.phase_factor = phase_factor
        self.old_hopping = old_hopping

    def __call__(self, i, j, *args):
        return np.dot(self.phase_factor(args), self.old_hopping)


class _StaticHoppingLeadToSystem:

    def __init__(self, phase_factor, old_hopping):

        self.phase_factor = phase_factor
        self.old_hopping = old_hopping

    def __call__(self, i, j, *args):
        return np.dot(self.old_hopping, self.phase_factor(args).conjugate())


def _make_time_dependent_hopping(phase, old_hopping, system_to_lead):
    """Return a time-dependent hopping function.
    """
    phase_factor = _PhaseFactor(phase)

    phase_signature = kwant._common.get_parameters(phase)

    if callable(old_hopping):

        old_hopping_signature = kwant._common.get_parameters(old_hopping)
        # strip off the two site parameters that are always present
        hopping_signature = old_hopping_signature[:2]
        hopping_parameter_signature = old_hopping_signature[2:]
        len_hopp = len(hopping_parameter_signature)

        # we arrange the arguments of new_hoppings such, that the first len_hopp
        # arguments are directly the parameters of the old hopping function
        # and for the remaining elements, we have a mapping array,
        # that mapps the input to the phase_factor function.
        # note that the args of the new_hopping function are unique,
        # whereas arguments might be present in both, the phase_factor
        # function as well as the old_hoppings.
        new_signature = hopping_parameter_signature + phase_signature
        unique, indices, inverse = np.unique(new_signature,
                                             return_index=True, return_inverse=True)
        inv_indices2 = np.argsort(indices)
        indices2 = np.argsort(inv_indices2)

        # mapping of new_hopping args to the phase function
        mapping = indices2[inverse[-len(phase_signature):]]

        new_hopping_signature = hopping_signature + tuple(unique[inv_indices2])

        if system_to_lead:
            new_hopping = _CallableHoppingSystemToLead(phase_factor, old_hopping, len_hopp, mapping)
        else:
            new_hopping = _CallableHoppingLeadToSystem(phase_factor, old_hopping, len_hopp, mapping)
    else:

        new_hopping_signature = ('_site0_', '_site1_') + phase_signature

        if system_to_lead:
            new_hopping = _StaticHoppingSystemToLead(phase_factor, old_hopping)
        else:
            new_hopping = _StaticHoppingLeadToSystem(phase_factor, old_hopping)

    _set_signature(new_hopping, new_hopping_signature)

    return new_hopping


def _make_block_tridiagonal(D, L, U, n):
    """Make a sparse block-tridiagonal matrix.

    Use ``D`` for the diagonal, ``L`` for the lower
    diagonal, and ``U`` for the upper diagonal. All
    three matrices must be square and have the same shape.

    Parameters
    ----------
    D, L, U : `~scipy.sparse.spmatrix`
        The diagonals, lower diagonals, and upper diagonals.
    n : int
        The number of copies of ``diag`` on the diagonal.

    Returns
    -------
    `~scipy.sparse.lil_matrix`
    """
    if D.shape != L.shape or D.shape != U.shape:
        raise ValueError('Diagonals, lower diagonals and upper diagonals must '
                         ' have the same shape.')
    if D.shape[0] != D.shape[1]:
        raise ValueError('Matrices must be square.')

    sz = D.shape[0]
    M = sp.block_diag([D] * n, format='lil')
    # TODO: convert this to use COO format, for efficiency
    rows = range(0, M.shape[0], sz)
    cols = range(sz, M.shape[0], sz)
    for i, j in zip(rows, cols):
        M[i:(i + sz), j:(j + sz)] = U
        M[j:(j + sz), i:(i + sz)] = L

    return M


def _make_square_matrix(M):
    """Make a square matrix from a rectangular one.

    The rectangular matrix should have shape (n, m)
    with n > m (i.e. upright rectangular). The returned
    matrix will be padded with zeros on the right and
    have shape (n, n).
    """
    if M.shape[0] > M.shape[1]:  # must make matrices square
        n = M.shape[0]
        s = M.shape[0] - M.shape[1]
        M = sp.hstack((M, sp.coo_matrix((n, s), dtype=M.dtype)))
    elif M.shape[0] < M.shape[1]:
        raise ValueError('Matrix must have shape (n, m) with n > m')
    return M


_inf = float('inf')


def _get_slice(syst, site):
    """Return a slice from the first orbital of this site to the next."""
    assert 0 <= site < syst.graph.num_nodes
    if syst.site_ranges is None:
        raise RuntimeError('Number of orbitals not defined.\n'
                           'Declare the number of orbitals using the '
                           '`norbs` keyword argument when constructing '
                           'the site families (lattices).')
    # Calculate the index of the run that contains the site.
    # column = np.asarray(syst.site_ranges)[:, 0]
    # run_idx = np.searchsorted(column, site, 'right') - 1
    # The `inf` is needed to avoid the fenceposting problem
    run_idx = bisect.bisect(syst.site_ranges, (site, _inf)) - 1
    # Calculate the slice.
    first_site, norbs, orb_offset = syst.site_ranges[run_idx]
    orb = orb_offset + (site - first_site) * norbs
    return slice(orb, orb + norbs)

# ----------


def _is_positive(a):
    """Check if a, respectively all elements of a, are positive."""
    try:
        return a.all() > 0
    except AttributeError:
        return a > 0


def _is_iterable(obj):
    try:
        _ = (e for e in obj)
        return True
    except TypeError:
        return False


def _make_iterable(variable):
    """If a variable is present, make it iterable"""
    if not _is_iterable(variable):
        return [variable]
    return variable


def _is_inside(x, xmin=None, xmax=None):
    """Check if xmin <= x <= xmax (elementwise if x is an array)."""
    x = np.array(x)
    if xmin is not None and xmax is not None:
        assert xmin <= xmax
        return (xmin <= x) & (x <= xmax)
    if xmin is not None and xmax is None:
        return xmin <= x
    if xmin is None and xmax is not None:
        return x <= xmax
    return np.ones(x.shape, dtype=bool)


def _get_interval(intervals, x):
    """From a list of intervals, get the ones where the point x is inside."""
    return [a for a in intervals if _is_inside(x, *a)]


def _get_unique_interval(intervals, x):
    """From a list of intervals, get the unique one where point x is inside."""
    interval = _get_interval(intervals, x)
    if len(interval) == 1:
        return interval[0]
    elif len(interval) > 1:
        raise ValueError('Interval is not unique')
    return None


def _make_grid(qmin, qmax, nq, gridtype):
    """Make a lin/log grid with nq points between (qmin, qmax)."""
    if qmax < qmin:
        raise ValueError('qmax={} must be larger qmin={}.'.format(qmax, qmin))
    if nq <= 0:
        raise ValueError('number nq={} must be larger zero.'.format(nq))
    if gridtype == 'log':
        if qmax <= 0:
            raise ValueError('qmax={} must be larger zero.'.format(qmax))
        if qmin <= 0:
            raise ValueError('qmin={} must be larger zero.'.format(qmin))
        return np.logspace(np.log10(qmin), np.log10(qmax), nq)
    if gridtype == 'lin':
        return np.linspace(qmin, qmax, nq)
    raise NotImplementedError('gridtype= {} not implemented'.format(gridtype))


class _Chain_1d(kwant.system.FiniteSystem):
    """Make a finite quasi 1D kwant system from a Hamiltonian.

    Parameters
    ----------
    n : int
        number of cells of the finite lead
    H : array like, shape (n*norbs, n*norbs)
        Hamiltonian matrix.
    norbs : int, optional
        Number of orbitals. Default=1
    leads : list of `kwant.system.InfiniteSystem`, optional
        List of Kwant leads to attach on the left and right of the system.
        Number of list elements can not be larger 2.
        If list is empty, no leads are attached. Default=None


    Returns
    -------
    syst : `kwant.system.FiniteSystem`
        Kwant system for the quasi 1D chain
    """

    def __init__(self, n, H, norbs=1, leads=None):

        assert _common.is_type(n, 'integer')
        assert _common.is_type(norbs, 'integer')

        if n < 0:
            raise ValueError('number of cells n={} must be positive.'.format(n))
        if norbs < 0:
            raise ValueError('number of orbitals norbs={} must be positive.'
                             .format(norbs))
        if H.shape != (n * norbs, n * norbs):
            raise ValueError('shape of the hamiltonian matrix={} does not '
                             'match the system size.'.format(H.shape))

        self.norbs = norbs
        self.H = H

        edges = [(i, i + 1) for i in range(n - 1)]
        edges += [elem[::-1] for elem in edges]

        g = kwant.graph.Graph()
        g.add_edges(edges)
        self.graph = g.compressed()

        self.site_ranges = [(0, norbs, 0), (n, 0, n * norbs)]

        if leads is not None:
            if len(leads) == 1:
                self.lead_interfaces = [np.array([0])]
            elif len(leads) == 2:
                self.lead_interfaces = [np.array([0]), np.array([n - 1])]
            elif len(leads) > 2:
                raise ValueError('number of leads={} is bigger than 2.'
                                 .format(len(leads)))
            self.leads = leads

    def hamiltonian(self, i, j, *args, **kwargs):
        i0 = i * self.norbs
        j0 = j * self.norbs
        return np.array([[self.H[i0 + ii, j0 + jj] for jj in range(self.norbs)]
                         for ii in range(self.norbs)], dtype=self.H.dtype)


def _finite_lead_hamiltonian(lead, num_cells, imaginary_potential=None,
                             params=None):
    """Extract the Hamiltonian matrix :math:`H` for a lead set to a
    finite lengths in presence of an imaginary potential.

    Parameters
    ----------
    lead : `kwant.system.InfiniteSystem`
        Kwant lead
    num_cells : int
        number of cells of the finite lead
    imaginary_potential : callable, optional
        imaginary potential function
    params : dict, optional
        Extra parameters for Hamiltonian value functions.

    Returns
    -------
    H : scipy.sparse.coo or scipy.sparse.lil matrix
        Hamiltonian matrix, shape (num_cells*norbs, num_cells*norbs)
        where norbs is the number of orbitals of the lead

    Notes
    -----
    If `imaginary_potential` is present, the Hamiltonian matrix
    will not be hermitian.
    """

    assert _common.is_type(num_cells, 'integer')
    if num_cells < 0:
        raise ValueError('number of cells num_cells={} must be positive.'
                         .format(num_cells))
    cell_norbs = _get_slice(lead, lead.cell_size).start

    simple_boundary = SimpleBoundary(num_cells=num_cells)
    H0 = simple_boundary(lead, params).hamiltonian.transpose()
    if imaginary_potential is None:
        return H0, cell_norbs
    if not callable(imaginary_potential):
        raise ValueError('Absorbing potential is not callable')
    bc_diags = [imaginary_potential(cell / num_cells) / num_cells
                for cell in range(num_cells)]
    S = sp.kron(sp.diags(bc_diags, offsets=0), sp.eye(cell_norbs))
    return H0 - 1j * S, cell_norbs


def _reflect(syst, energy, params=None):
    """Calculate the reflection amplitude :math:`r` for an absorbing lead."""
    assert _common.is_type(energy, 'real_number')
    smatrix = kwant.smatrix(syst, energy, check_hermiticity=False,
                            params=params).submatrix(0, 0)
    return np.array([np.linalg.norm(smatrix[mode, :]) for mode
                     in range(smatrix.shape[0])])


def _modes(lead, energy, params=None):
    """Calculate the momentum and velocities of foreward propagating modes."""
    assert _common.is_type(energy, 'real_number')
    modes = lead.modes(energy, params=params)[0]
    nmodes = modes.momenta.size // 2
    return modes.momenta[nmodes:], modes.velocities[nmodes:]


class AbsorbingReflectionSolver:

    """Calculate the reflection amplitude :math:`r` for a lead
    with imaginary absorbing potential.

    Examples
    --------
    >>> num_cells = 100
    >>> def my_imaginary_potential(x):
            return 50 * x**4
    >>> reflect = AnalyzeReflection(lead, num_cells, my_imaginary_potential)
    >>> energies = numpy.linspace(0, 0.5, 101)
    >>> r = reflect(energies)


    Notes
    -------
    An instance of this class can be called like a method to calculate
    the reflection amplitude :math:`r` of all open modes at a given energy.
    The reflection amplitude is calculated by a static kwant calculation.
    """

    def __init__(self, lead, num_cells, imaginary_potential, params=None):
        """
        Parameters
        ----------
        lead : `kwant.system.InfiniteSystem`
            Kwant lead, for which the reflection should be calculated.
        num_cells : int
            number of cells for a finite lead taken into account
        imaginary_potential : callable
            absorbing potential function
        params : dict, optional
           Extra parameters for Hamiltonian value functions of ``lead``.

        """

        if not isinstance(lead, kwant.system.InfiniteSystem):
            raise TypeError('lead must be an instance of `InfiniteSystem`')

        self.lead = lead
        self.params = params

        H, cell_norbs = _finite_lead_hamiltonian(lead, num_cells,
                                                 imaginary_potential,
                                                 params=params)
        self.syst = _Chain_1d(num_cells, H, norbs=cell_norbs, leads=[lead])

    def __call__(self, energies):
        """ Calculates the reflection amplitude :math:`r` for all open modes
        at the given energies.

        Parameters
        ----------
        energies : float or list of floats shape (ne, )
            energies for which the reflection should be calculated

        Returns
        -------
        reflect : list, shape(ne, ) or numpy array
            Reflection amplitude, ordering similar to `energies`.
            For each `energy` element in `energies`, the corresponding
            `reflect` list element is a numpy array of shape (nmodes, ), where
            `reflect` the number of open modes at this energy. If `energies`
            was a float, the reflection of the the open modes are returned
            directly as a numpy array of shape (nmodes, ).
        momenta : list, shape(ne, ) or numpy array
            Momenta of the mode, ordering similar to `energies`.
            For each `energy` element in `energies`, the corresponding
            `momenta` list element is a numpy array of shape (nmodes, ),
            where nmodes is the number of open modes at this energy.
            If `energies` was a float, the momenta of the the open modes are
            returned directly as a numpy array of shape (nmodes, ).
        velocities : list, shape(ne, ) or numpy array
            Velocities of the mode, ordering similar to `energies`.
            For each `energy` element in `energies`, the corresponding
            `velocities` list element is a numpy array of shape (nmodes, ),
            where nmodes is the number of open modes at this energy.
            If `energies` was a float, the velocities of the the open modes are
            returned directly as a numpy array of shape (nmodes, ).


        Notes
        -------
        If no modes are open at a given energy, the corresponding lists are
        empty. If the energy matches exactly the band openings, the following
        error might occur: 'ValueError: Input a needs to be a square matrix.'
        """

        if _is_iterable(energies):
            reflect = [_reflect(self.syst, energy, params=self.params)
                       for energy in energies]
            momenta, velocities = zip(*[_modes(self.lead, energy,
                                               params=self.params)
                                        for energy in energies])
        else:
            reflect = _reflect(self.syst, energies, params=self.params)
            momenta, velocities = _modes(self.lead, energies,
                                         params=self.params)

        return reflect, momenta, velocities


class AnalyzeReflection:
    """ Analyze the the reflection for a lead."""

    def __init__(self, lead, num_cells, imaginary_potential,
                 spectrum=None, params=None):
        """
        Parameters
        ----------
        lead : `kwant.system.InfiniteSystem`
            Kwant lead, for which the reflection should be calculated.
        num_cells : int
            number of cells for a finite lead taken into account
        imaginary_potential : callable
            absorbing potential function
        spectrum : `BandSketching`, optional
            `kwant.physics.BandSketching` instance of the lead.
            This is mainly for performance.
            If not present, it will be calculated on the fly from `lead`.
        params : dict, optional
            Extra parameters for Hamiltonian value functions of ``lead``.

        """

        self.reflection_solver = AbsorbingReflectionSolver(lead, num_cells,
                                                           imaginary_potential,
                                                           params=params)

        if spectrum is None:
            self.spectrum = kwantspectrum.spectrum(lead, params=params)

    def __call__(self, k, band):
        """Calculates the reflection amplitude :math:`r` for a mode with
        momentum :math:`k` for a given band.

        Parameters
        ----------
        k : int or float, scalar or array, shape (n, )
            Momentum values where the reflection should be calculated.
        band : int
            band index

        Returns
        -------
        reflect : numpy array, shape(n, )
            Reflection amplitude
        energies : numpy array, shape(n, )
            Mode energy
        vel : numpy array, shape(n, )
            Mode velocity
        """

        energies = self.spectrum(k, band)
        rm, km, vm = self.reflection_solver(energies)

        # TODO: what happens if, due to numeric. inacc, rm is empty ?

        # if several modes are open at a given energy, get the mode
        # that corresponds to the band at given momentum k
        try:  # expect k to be iterable
            pos = [np.argmin(np.abs(ki - k0)) for ki, k0 in zip(km, k)]
            ref = np.array([ri[i] for ri, i in zip(rm, pos)])
            vel = np.array([vi[i] for vi, i in zip(vm, pos)])
        except TypeError:  # k is a scalar
            pos = np.argmin(np.abs(km - k))
            ref = rm[pos]
            vel = vm[pos]

        return ref, energies, vel

    def around_extremum(self, kmin, kmax, band, nq=20, dq=0.001, gridtype='log'):
        """Calculate the reflection amplitude :math:`r` within the momentum
        interval ``[kmin, kmax]`` around a local extremum of the dispersion.

        Parameters
        ----------
        kmin, kmax : float
            momentum interval ``[kmin, kmax]`` including the a single local
            dispersion minimum or maximum located at ``k0``.
        band : int
            band index
        nq : int, optional
            number of sample gridpoints between ``[kmin, k0 - dq]`` and between
            ``[k0 + dq, kmax]``.
            The total number of sample points is ``2 * nq``.
        dq : float, optional
            Offset around extrema located at momentum ``k0``.
        gridtype : string, optional
           `lin` : use a linear grid between ``[kmin, k0]`` and ``[k0, kmax]``

           `log` : use a log grid between ``[kmin, k0]`` and ``[k0, kmax]``.
                   The log-grid is choosen dense around ``k0``.

        Returns
        -------
        reflect : numpy array, shape(``2 * nq``, )
            Reflection amplitude :math:`r`
        energies : numpy array, shape(``2 * nq``, )
            Mode energies
        vel : numpy array, shape(``2 * nq``, )
            mode velocities
        k : numpy array, shape(``2 * nq``, )
            mode momenta
        e0 : float
            dispersion energy at the local minimum or maximum
        k0 : float
            momentum of the local minimum or maximum

        Notes
        -------
        The local extremum of the dispersion ``k0`` must be inside the
        interval  ``[kmin, kmax]``, such that  ``kmin <= k0 <= kmax``.
        Otherwise, a ``ValueError`` is thrown.
        """
        # get minimum/maximum position at k0
        k0, e0 = _get_extremum(kmin, kmax, band, self.spectrum)
        # calc reflection on left/right intervals: [kmin, k0-dq], [k0+dq, kmax]
        kl = k0 - _make_grid(dq, k0 - kmin, nq, gridtype)[::-1]
        kr = k0 + _make_grid(dq, kmax - k0, nq, gridtype)
        k = np.concatenate((kl, kr), axis=0)
        refl, ener, vel = self.__call__(k, band)
        return refl, ener, vel, k, e0, k0


class AnalyzeReflectionMonomial:
    """ Analyze the the reflection for a lead for the special case of a
    monomial absorbing potential.

    Notes
    -------
    The reflection amplitude is calculated with analytic expressions
    derived in [1]_.
    Good agreement with the exact numerical result (via `AnalyzeReflection`)
    can only be expected if :math:`k * l >> 1`.
    The length :math:`l` corresponds to `num_cells` and the momentum
    :math:`k` is the distance from a local extremum of the spectrum.

    .. [1] `J. Weston and X. Waintal, Phys. Rev. B 93, 134506 (2016)
       <https://arxiv.org/abs/1510.05967>`_
    """

    def __init__(self, lead, num_cells, strength, degree, params=None):
        """
        Parameters
        ----------
        lead : `kwant.system.InfiniteSystem` or `kwantspectrum.BandSketching`
            Kwant lead or `BandSketching` instance of the lead,
            for which the reflection should be calculated.
        num_cells : int
            number of cells for a finite lead taken into account
        degree : int
            order of the moniminal absorbing boundary potential.
        strength : float
            The strength of the boundary conditions. Formally
            this is the area underneath the monomial curve.
        params : dict, optional
            Extra parameters for Hamiltonian value functions of ``lead``.
            ``params`` is only considered if ``lead`` is an instance of
            `kwant.system.InfiniteSystem`.
        """
        if isinstance(lead, kwant.system.InfiniteSystem):
            self.spectrum = kwantspectrum.spectrum(lead, params=params)
        else:
            self.spectrum = lead

        if num_cells <= 0:
            raise ValueError('Number of unit cells must be positive.')

        self._reflect = ft.partial(_monomial_reflect, length_absorb=num_cells,
                                   strength=strength, degree=degree)

    def around_extremum(self, kmin, kmax, band, nq=20, dq=0.001, gridtype='log'):
        """Calculate the reflection amplitude :math:`r` within the momentum
        interval ``[kmin, kmax]`` around a local extremum of the dispersion.

        Parameters
        ----------
        kmin, kmax : float
            momentum interval ``[kmin, kmax]`` including the a single local
            dispersion minimum or maximum located at ``k0``.
        band : int
            band index
        nq : int, optional
            number of sample gridpoints between ``[kmin, k0 - dq]`` and between
            ``[k0 + dq, kmax]``.
            The total number of sample points is ``2 * nq``.
        dq : float, optional
            Offset around extrema located at momentum ``k0``.
        gridtype : string, optional
           `lin` : use a linear grid between ``[kmin, k0]`` and ``[k0, kmax]``

           `log` : use a log grid between ``[kmin, k0]`` and ``[k0, kmax]``.
                   The log-grid is choosen dense around ``k0``.

        Returns
        -------
        reflect : numpy array, shape(``2 * nq``, )
            Reflection amplitude :math:`r`
        energies : numpy array, shape(``2 * nq``, )
            Mode energies
        vel : numpy array, shape(``2 * nq``, )
            mode velocities
        k : numpy array, shape(``2 * nq``, )
            mode momenta
        e0 : float
            dispersion energy at the local minimum or maximum
        k0 : float
            momentum of the local minimum or maximum

        Notes
        -------
        The local extremum of the dispersion ``k0`` must be inside the
        interval  ``[kmin, kmax]``, such that  ``kmin <= k0 <= kmax``.
        Otherwise, a ``ValueError`` is thrown.
        """
        # get minimum/maximum position at k0
        k0, e0 = _get_extremum(kmin, kmax, band, self.spectrum)
        # calc reflection on left/right intervals: [kmin, k0-dq], [k0+dq, kmax]
        kl = k0 - _make_grid(dq, k0 - kmin, nq, gridtype)[::-1]
        kr = k0 + _make_grid(dq, kmax - k0, nq, gridtype)
        k = np.concatenate((kl, kr), axis=0)
        ener = self.spectrum(k, band)
        vel = self.spectrum(k, band, derivative_order=1)
        refl = self._reflect(np.abs(ener - e0), np.abs(k - k0))
        return refl, ener, vel, k, e0, k0

# -------
# analytical routines for reflection
# -------


def _strength_opti(e, k, length_absorb, degree):
    r""" Monominal absorbing boundary strength :math:`A` for which the absolute
    reflection amplitude :math:`r_\Sigma` is minimal.

    Returns
    -------
    strength : float
        Optimal strength `A` of the monomial absorbing potential.

    Notes
    -------
    The analytical expression can be obtained from
        d r_\Sigma / d A = 0 in Eq. (34)
        J. Weston and X. Waintal, Physical Review B 93, 134506 (2016).
    Note that the formula Eq. (34) is wrong, but we have to replace
        (n-1) -> (n-1)!
    """
    assert _is_positive(e)
    assert _is_positive(k)
    assert _is_positive(length_absorb)
    assert _is_positive(degree)
    degplus = degree + 1
    num = degree * degplus * np.math.factorial(degree - 1)
    denom = 2 * (2 * length_absorb * k) ** degplus
    return - e / k * np.log(num / denom)


def _monomial_reflect(e, k, length_absorb, strength, degree):
    """Reflection amplitude 'r' for the monomial potential.

    Returns
    -------
    refl : float
        reflection amplitude `r`, the value ranges between 0 and 1

    Notes
    -------
    The analytical expression corresponds to Eq. (34)
         J. Weston and X. Waintal, Physical Review B 93, 134506 (2016).
    Note that the formula Eq. (34) is wrong, but we have to replace
        (n-1) -> (n-1)!
    We also limit 'r' to values not larger then one.
    """
    assert _is_positive(e)
    assert _is_positive(k)
    assert _is_positive(length_absorb)
    assert _is_positive(strength)
    assert _is_positive(degree)
    nm1fac = np.math.factorial((degree - 1))
    num = degree * (degree + 1) * nm1fac
    denom = 4 * length_absorb * (2 * length_absorb * k) ** degree
    refl = np.abs(np.exp(-strength * k / e) + strength * num / (e * denom))
    return np.where(refl < 1, refl, 1)


def _low_energy_reflect(e, k, length_absorb, degree, strength):
    """Reflection amplitude 'r' for the low energy mode in monomial potential.

    Returns
    -------
    refl : float
        reflection amplitude `r`, the value ranges between 0 and 1
    """
    assert _is_positive(e)
    assert _is_positive(k)
    assert _is_positive(length_absorb)
    assert _is_positive(degree)
    assert _is_positive(strength)
    nm1fac = np.math.factorial(degree - 1)
    degplus = degree + 1
    nt = degree * degplus * nm1fac / 2
    at = strength * k / e
    denom = 2 * k * length_absorb
    refl = np.abs(np.exp(-at) + nt * at * (1 / denom)**degplus)
    return np.where(refl < 1, refl, 1)


def _monomial_absorbing_potential(strength, degree):
    """Monomial potential function used for absorbing boundaries.
    Form similar to equation (33) in:
    J. Weston and X. Waintal, Physical Review B 93, 134506 (2016).
    """
    assert _is_positive(degree)

    def pot(x, n, a):
        return (n + 1) * a * x**n
    return ft.partial(pot, a=strength, n=degree)


def _max_buffer_velocity(num_buffer_cells, tmax):
    """returns the maximal velocity such that a mode stays in buffer"""
    if num_buffer_cells < 0:
        raise ValueError('Cannot have a negative number of buffer cells.')
    if tmax <= 0:
        raise ValueError('Maximum time must be positive.')
    return 2 * num_buffer_cells / tmax  # factor 2 due to back and forth


def _optimal_split(degree):
    r"""Optimal splitting `x` in buffer/absorbing zone for the low energy modes
    Formula obtained from d r_\Sigma / d x = 0 in low energy approximation

    Definition
    ----------

        length buffer zone (`num_buffer_cells`) = x * length
        length absorbing zone (`num_cells`) = (1-x) * length
        length is total length of additional lead cells
        (`num_cells + num_buffer_cells`)

    Returns
    -------
    x : float
        optimal splitting parameter
    """
    if degree < 0:
        raise ValueError('Degree={} must be positive'.format(degree))
    return (2 + degree) / (3 + 2 * degree)

# -------
# analytical routines for reflection - end
# -------


# -------
# routines to do some band structure gymnastics
# TODO:
# rewrite spectrum more flexible to include most of the
# logic in spectrum
# if the spectrum is truncted by emin/emax, maximal curvature and velocity
# points might not exist. handle this case correctly.
# -------


def _get_extremum(kmin, kmax, band, spectrum):
    """Get the unique minima/maxima of the band at k0 that satisfies
    kmin <= k0 <= kmax.

    Returns
    -------
    k0 : float, momentum at the extremum
    e0 : float, energy at the extremum
    """
    zeros = spectrum.intersect(f=0, band=band, derivative_order=1, kmin=kmin, kmax=kmax)
    if len(zeros) != 1:  # test for existance and uniqueness
        raise ValueError('(kmin, kmax) interval must be choosen such that '
                         'exactly one extremum of the dispersion at k0 '
                         'lies within. Here k0={}.'.format(zeros))
    return zeros[0], spectrum(zeros[0], band)


def _max_curvature_point(spectrum, emin=None, emax=None):
    """Find the maximal velocity, respectively curvature of the spectrum.

    Parameters
    ----------
    emin, emax: float, optional
        if not `None`, we restrict to the zone bounded by [emin, emax].
    Returns
    -------
    g0 : float
        maximal absolute value of the curvature (`g0` >= 0) in
        a local minimum or maximum located at energy `e0`, momentum `k0`
        and band with index `band`
    """
    def energy_extrema(band):
        """Return an array with momenta, where velocity has local extrema"""
        # unconstrained case: velocity derivative is zero
        zeros = spectrum.intersect(f=0, band=band, derivative_order=1)
        # filter points with energies in [emin, emax]
        if _common.is_not_empty(zeros):
            energies = spectrum(zeros, band)
            inside = _is_inside(energies, emin, emax)
            zeros = zeros[inside]
        return zeros

    # find the largest curvature on all extremum points
    d2E_b = [0] * spectrum.nbands
    k_b = [0] * spectrum.nbands
    for band in range(spectrum.nbands):
        # get all momenta where energy is maximal
        k_extrema = energy_extrema(band)
        # now find the maximal curvature, if several minima are present
        if _common.is_not_empty(k_extrema):
            d2E = spectrum(k_extrema, band, derivative_order=2)
            max_id = np.argmax(np.abs(d2E))
            d2E_b[band] = d2E[max_id]
            k_b[band] = k_extrema[max_id]

    # find the largest curvature from all bands
    max_band = np.argmax(np.abs(d2E_b))
    gmax = d2E_b[max_band]
    if gmax == 0:
        logger.warning('could not find maximal curvature value for '
                       'spectrum within emin={}, emax={}'. format(emin, emax))
        raise RuntimeError('unable to find max curvature point of the spectrum')
    kmax = k_b[max_band]
    energy_max = spectrum(kmax, max_band)
    return kmax, energy_max, max_band, gmax


def _max_curvature(spectrum, emin=None, emax=None):
    """Find the (local) dispersion minima/maxima with the largest curvature.
    """

    k0, e0, band, g0 = _max_curvature_point(spectrum, emin=emin, emax=emax)

    kmin = 2 * spectrum.kmin
    kmax = 2 * spectrum.kmax

    # get the momentum interval with positive velocity sourrounding k0
    # and set k1 to the momentum of either the left or right interval border,
    # such that the distance to k0 is minimal
    intervals_e = spectrum.intervals(band, lower=emin, upper=emax,
                                     kmin=kmin, kmax=kmax)
    epsilon = 1E-6  # second derivative might not become exactly zero
    if np.sign(g0) > 0:
        intervals_g = spectrum.intervals(band, lower=epsilon, derivative_order=2,
                                         kmin=kmin, kmax=kmax)
    else:
        intervals_g = spectrum.intervals(band, upper=-epsilon, derivative_order=2,
                                         kmin=kmin, kmax=kmax)
    intervals = kwantspectrum.intersect_intervals(intervals_e, intervals_g)
    interval = _get_unique_interval(intervals, k0)
    q = np.min(np.abs(np.array(interval) - k0))

    logger.debug('local min/max with hightest curvature g: g={}, k={}, energy={},'
                 'band={}, neighbor extrema with smallest momentum distance at '
                 'relative momentum q={}'.format(g0, k0, e0, band, q))
    return k0, e0, band, q, g0


def _max_velocity_point(spectrum, emin=None, emax=None):
    """Find the maximal velocity of the spectrum.

    Parameters
    ----------
    emin, emax: float, optional
        if not `None`, we restrict to the zone bounded by [emin, emax].
    Returns
    -------
    kmax : float
        mometum value at maximum
    energy_max : float
        energy value at maximum
    max_band : int
        band with maximum velocity
    vmax : float
        maximum velocity
    """
    def velocity_maxima(band):
        """Return an array with momenta, where velocity has local extrema"""
        # unconstrained case: velocity derivative is zero
        zeros = spectrum.intersect(f=0, band=band, derivative_order=2)
        # if noise prevents to find E''(k) ==0, we look for E'(k) > 0 intervals
        if not _common.is_not_empty(zeros):
            v_pos = spectrum.intervals(band, lower=0, derivative_order=1)
            zeros = np.array(v_pos).flatten()
        # filter points with energies in [emin, emax]
        if _common.is_not_empty(zeros):
            energies = spectrum(zeros, band)
            inside = _is_inside(energies, emin, emax)
            zeros = zeros[inside]
        # velocity maximization with energy constrained:
        # if velocity maxima are outside the energy range
        # the band cuts at emin/emax are local extrema
        if emin is not None:
            inters = spectrum.intersect(f=emin, band=band)
            zeros = np.append(zeros, inters)
        if emax is not None:
            inters = spectrum.intersect(f=emax, band=band)
            zeros = np.append(zeros, inters)
        return zeros

    # find the largest velocity per band
    dE_b = [-np.inf] * spectrum.nbands
    k_b = [-np.inf] * spectrum.nbands
    for band in range(spectrum.nbands):
        # get all momenta where velocity is maximal
        momenta_vmax = velocity_maxima(band)
        # now find the maximal velocity, if several minima are present
        if _common.is_not_empty(momenta_vmax):
            dE = spectrum(momenta_vmax, band, derivative_order=1)
            max_id = np.argmax(dE)
            dE_b[band] = dE[max_id]
            k_b[band] = momenta_vmax[max_id]

    # find the largest velocity from all bands
    max_band = np.argmax(dE_b)
    vmax = dE_b[max_band]
    if vmax == -np.inf:
        logger.warning('could not find maximum velocity value for '
                       'spectrum within emin={}, emax={}'.format(emin, emax))
        raise RuntimeError('unable to find max velocity point of the spectrum')
    kmax = k_b[max_band]
    energy_max = spectrum(kmax, max_band)
    return kmax, energy_max, max_band, vmax


def _max_velocity(spectrum, emin=None, emax=None):
    """Find the highest velocity of the spectrum and the minimum
    distance to the sourrounding local extrema

    Returns
    -------
    k0 : float
        momentum of the maximum positive velocity point
    e0 : float
        E(k0, band) (energy at the maximum positive velocity point)
    band : int
        band index of the band with vmax
    eq : float
        eq = | E(k0) - E(k1) | (relative energy distance)
    q : float
        q = |k0 - k1| (minimum relative distance to surounding minimum/maximum)
    vmax : float
        vmax = E'(k0, band) > 0 (maximal positive velocity of the spectrum)
    """

    # get max velocity point
    k0, e0, band, vmax = _max_velocity_point(spectrum, emin=emin, emax=emax)
    kmin = 2 * spectrum.kmin
    kmax = 2 * spectrum.kmax

    # get the momentum interval with positive velocity sourrounding k0
    # and set k1 to the momentum of either the left or right interval border,
    # such that the distance to k0 is minimal
    intervals_e = spectrum.intervals(band, lower=emin, upper=emax,
                                     kmin=kmin, kmax=kmax)
    intervals_v = spectrum.intervals(band, lower=0, derivative_order=1,
                                     kmin=kmin, kmax=kmax)
    intervals = kwantspectrum.intersect_intervals(intervals_e, intervals_v)

    k_left, k_right = _get_unique_interval(intervals, k0)

    # if k0 is a velocity maxima with E''(k0)=0, we search a k1 with
    # E'(k1) = 0, such that |k0 - k1| is minimal. by construction, k1
    # corresponds either to k_left or k_right.
    # on the other hand if k0 is not a velocity maxima (E''(k0)/=0),
    # then, by construction, k0 will be either k_left or k_right.
    # we set k1 to the opposite momentum, since this is the nearest possible
    # energy extremum.
    if _common.is_zero(k0 - k_left):  # k0 = k_left < k_right, E''(k0) /= 0
        k1 = k_right
    elif _common.is_zero(k0 - k_right):  # k_left < k_right = k0, E''(k0) /= 0
        k1 = k_left
    else:  # k0 with k_left < k0 < k_right
        k1 = k_left if abs(k0 - k_left) < abs(k0 - k_right) else k_right

    # relative absolute energy and momentum of the maximal velocity point
    # from local surounding extremum
    e1 = spectrum(k1, band)
    q = np.abs(k1 - k0)
    eq = np.abs(e1 - e0)

    logger.debug('highest velocity={} at k={}, energy={}, '
                 'band={}, neighbor extrema at k={}, energy={}'
                 . format(vmax, k0, e0, band, k1, e1))

    return k0, e0, band, q, eq, vmax


def _fast_mode(spectrum, emin, emax):
    """Fast (high) energy mode of a spectrum

    Returns
    -------
    eq : float
        eq = | E(kmax) - E(k0) | (relative energy distance)
    vmax : float
        vmax = E'(kmax) > 0 (maximal positive velocity of the spectrum)
    q : float
        q = |kmax - k0| (minimum relative distance to surounding minimum/maximum)

    Notes
    -----
    k0 is the momentum of either the left or right extremum around kmax
    such that q, the relative distance, is minimal
    """
    *_, q, eq, vmax = _max_velocity(spectrum, emin, emax)
    return eq, vmax, q


def _slow_mode(spectrum, emin, emax):
    r"""Slow (low) energy mode of a spectrum

    Returns
    -------
    disp : callable
        disp = |E(k)|, with E(0) = 0 (energy dispersion parametrization)
    vel : callable
        vel = |E'(k)|, with E'(0) = 0 (velocity function parametrization)
    q : float
        maximum momentum (k) argument k \in [-q, q] of `disp` and `vel`

    Notes
    -----
    The dispersion function disp parametrizes the spectrum around the local
    extremum with the hightest curvature. The maximum momentum q
    is chosen such that disp is a strong monotonous function for k \in [0, q]
    and k \in [0, -q].
    """
    k0, e0, band, q, *_ = _max_curvature(spectrum, emin, emax)
    v0 = spectrum(k0, band, derivative_order=1)

    def disp(k):
        # in quadratic approximation E(k) = g0 * k**2
        return np.abs(spectrum(k0 + k, band) - e0)

    def vel(k):
        # in quadratic approximation v(k) = 2 * g0 * k
        return np.abs(spectrum(k0 + k, band, derivative_order=1) - v0)

    return disp, vel, q


# -------
# end of band structure analysis
# ----


def _maximal_buffer_momentum(tmax, len_buffer, vel_func, kmax):
    """Find the maximal momentum of a mode such that it stays into the buffer

    Parameters
    ----------
    tmax : int or float
        Maximal time for a tkwant simulation to run. Must be positive > 0.
    len_buffer : float
        length of the buffer zone (`num_buffer_cells`)
    vel_func : callable
        velocity function `v(k)` with k in [0, kmax]
    kmax : float
        Maximal value of the momentum argument of the velocity function

    Returns
    -------
    k : float
        maximal momentum that a mode with velocity function `v(k)` can have
        such that it stays inside the buffer. k has values between 0 and kmax.

    Notes
    -------
    mathematical description of the routine:

        vmax = 2 * len_buffer / tmax > 0 (maximal velocity of a mode
                                              such that it stays into buffer)
        we require v(k) to be monotonically increasing (for uniquness)
        and that kmax > 0, such that v(0) < v(kmax)

        - if v(0) <= vmax <= v(kmax):
            find the unique solution k in [0, kmax] such that v(k) = vmax

        - if vmax > v(kmax) (buffer zone larger then fastest mode)
            we return kmax

     modes with momenta in [0, k] stay into the buffer zone
     modes with momenta in [k, kmax] will escape buffer zone
     (and have to be captured by the absorbing zone)
    """

    # the checks should be save even if vel_func
    # has some numerical inaccuracy
    assert len_buffer > 0
    assert tmax > 0
    assert kmax > 0
    assert vel_func(0) < vel_func(kmax)

    def func(k):
        return vel_func(k) - 2 * len_buffer / tmax

    assert func(0) < 0
    if func(kmax) < 0:
        return kmax
    return brentq(func, 0, kmax)


def _reflect_slow_mode(slow_mode, tmax, strength, degree, length):
    """Estimate the reflection of the slow mode in the lead (buffer+aborb)
    in monomial approx.

    Parameters
    ----------
    slow_mode : tuple
        parametrization of the slowest, low energy mode
    tmax : int or float
        Maximal time for a tkwant simulation to run. Must be positive > 0.
    strength : float
        The strength of the boundary conditions. Formally
        this is the area underneath the monomial curve.
    degree : int
        order of the monominal absorbing boundary potential.
        Must be positive > 0.
    length : float
        total length of all additional lead cells
        (num_cells+num_buffer_cells)

    Returns
    -------
    refl : float
        reflection amplitude `r`, the value ranges between 0 and 1

    Notes
    -------
    If the routine encounters numerical problems, it returns
    a reflection coefficient of one.
    """

    x = _optimal_split(degree)
    length_buffer = x * length  # (corresponds to `num_buffer_cells`)
    length_absorb = (1 - x) * length  # (corresponds to `num_cells`)

    # Get maximal momentum `k` of slow mode such that it stays into the buffer
    energy_func, vel_func, kmax = slow_mode
    if not callable(energy_func):
        raise ValueError('Energy function is not callable')
    if not callable(vel_func):
        raise ValueError('Velocity function is not callable')
    try:
        k = _maximal_buffer_momentum(tmax, length_buffer, vel_func, kmax)
    except Exception:  # numerical accuracy can spoil the inversion
        logger.warning('k-inversion problem: {}, {}, {}'
                       .format(tmax, length_buffer, kmax))
        return 1

    # Return the reflection of the absorbing zone
    try:
        return _low_energy_reflect(energy_func(k), k, length_absorb, degree,
                                   strength)
    except Exception:
        logger.warning('low energy reflection problem: {}, {}, {}, {}'
                       .format(energy_func(k), k, length_absorb, strength))
        return 1


def _optimize_length(fast_mode, slow_mode, tmax, refl_max, strength, degree,
                     length_init, lmin=1):
    """Find the minimal possible length for an absorbing lead, such that
     `r < refl_max`, while keeping the strength fixed.

    Parameters
    ----------
    fast_mode : tuple
        parameters characterizing the fastest mode of the spectrum
    slow_mode : tuple
        parametrization of the slowest, low energy mode
    tmax : int or float
        Maximal time for a tkwant simulation to run. Must be positive > 0.
    refl_max : float
        maximal allowed reflection amplitude :math:`r`, must be in (0, 1)
        interval.
    strength : float
        The strength of the boundary conditions. Formally
        this is the area underneath the monomial curve.
    degree : int
        order of the monominal absorbing boundary potential.
        Must be positive > 0.
    length_init : float
        initial total length of all additional lead cells
        (num_cells+num_buffer_cells)
    lmin : float, optional
        minimal length to perform numerical root finding

    Returns
    -------
    length : float
        new estimate for the total length of the additional lead cells
        (`num_cells + num_buffer_cells`)
        The returned value of length in inside the inteval [lmin, length_init]
    success : bool
        `True`: new smaller length found: `length < length_init`
         such that : `r < refl_max`.
        `False`: not able to find a smaller length such that:
        `length < length_init` and `r < refl_max`.
        `r` is the total reflection amplitude of the lead
    """

    if refl_max <= 0:
        raise ValueError('Reflection coefficint={} must be positive'
                         .format(refl_max))

    assert 0 <= lmin <= length_init
    momentum, energy, _ = fast_mode
    xabs = (1 - _optimal_split(degree))

    def func(length):  # reflection criterion `r < refl_max`
        length_absorb = length * xabs
        r1 = _monomial_reflect(energy, momentum, length_absorb, strength, degree)
        r2 = _reflect_slow_mode(slow_mode, tmax, strength, degree, length)
        return max(r1, r2) - refl_max

    # try to find a smaller length
    try:
        length = brentq(func, lmin, length_init)
        success = True
    except Exception:
        length = length_init
        success = False
    return length, success


def _optimize_strength(fast_mode, refl_max, degree, length_init):
    """Find an appropriate strength of the monomial absorbing potential,
    such that such that `r < refl_max`, while keeping the length fixed.

    Parameters
    ----------
    fast_mode : tuple
        parameters characterizing the fastest mode of the spectrum
    refl_max : float
        maximal allowed reflection amplitude :math:`r`, must be in (0, 1)
        interval.
    degree : int
        order of the monominal absorbing boundary potential.
        Must be positive > 0.
    length_init : float
        initial total length of all additional lead cells
        (`num_cells + num_buffer_cells`)

    Returns
    -------
    strength : float
        The strength of the boundary conditions. Formally
        this is the area underneath the monomial curve.
        `strength` is only meaningful if `success` is `True`
    success : bool
        `True`: strength found such that: `r < refl_max`
        `False`: not able to find strength such that:
        `r < refl_max`
        `r` is the total reflection amplitude of the lead
    """

    # the total lead reflection `r` has absorption and reflection contribution.
    # a large strength lowers absorption but increases reflection and
    # vice versa. the reflection contribution (but not the absorption contrib.)
    # depends however also on the length over which the potential is smeared
    # (`num_cells`) and can be lowered if we increase the length.
    # the goal of this routine is to find a strength sufficiently large that
    # the fast mode is absorbed. we assume that the fast mode needs
    # the largest strength A. the strength should stay as small as possible
    # however to keep the reflection contribution, especially from the slow
    # modes, small, in order to choose the length small as well.

    # TODO: can we tune here ?
    # we could try to lower the strength iteratively, this
    # is just a one-shot try and error.
    if refl_max <= 0:
        raise ValueError('Reflection coefficint={} must be positive'.
                         format(refl_max))

    momentum, energy, _ = fast_mode

    length_absorb = length_init * (1 - _optimal_split(degree))
    strength = _strength_opti(energy, momentum, length_absorb, degree)
    success = _monomial_reflect(energy, momentum, length_absorb, strength,
                                degree) <= refl_max
    return strength, success


def _new_length_estimate(fast_mode, slow_mode, tmax, refl_max, degree,
                         length_init):
    """Try to find an appropriate strength/length combination for the monomial
    absorbing potential such that the new length < length_init

    Parameters
    ----------
    fast_mode : tuple
        parameters characterizing the fastest mode of the spectrum
    slow_mode : tuple
        parametrization of the slowest, low energy mode
    tmax : int or float
        Maximal time for a tkwant simulation to run. Must be positive > 0.
    refl_max : float
        maximal allowed reflection amplitude :math:`r`, must be in (0, 1)
        interval.
    degree : int
        order of the monominal absorbing boundary potential.
        Must be positive > 0.
    length_init : float
        initial total length of all additional lead cells
        (num_cells+num_buffer_cells)

    Returns
    -------
    length : float
        new estimate for the total length of the additional lead cells
        (`num_cells + num_buffer_cells`)
    strength : float
        The strength of the boundary conditions. Formally
        this is the area underneath the monomial curve.
        `strength` is only meaningful if `success` is `True`
    success : bool
        `True`: new smaller length found: `length < length_init`
        `False`: not able to find a smaller length such that:
        `length < length_init`
    """

    # use the fast mode to find a strength sufficiently large
    # keep the length fixed to the initial length `length_init`
    strength, success = _optimize_strength(fast_mode, refl_max, degree,
                                           length_init)

    # try to find a smaller length such that both fast and slow modes
    # are reflected less then allowd by `refl_max`.
    # the strength is kept constant.
    if success:
        length, success = _optimize_length(fast_mode, slow_mode, tmax,
                                           refl_max, strength, degree,
                                           length_init)
    else:
        length = length_init

    return length, strength, success


def _monomial_parameter_estimate(spectrum, tmax, refl_max, degree,
                                 emin=None, emax=None, eps=1E-4):
    """Estimate (optimal) parameters for the absorbing potential in monomial
    approximation.

    Parameters
    ----------
    spectrum : `BandSketching`
        `kwant.physics.BandSketching` instance of the lead.
    tmax : int or float
        Maximal time for a tkwant simulation to run. Must be positive > 0.
    refl_max : float
        maximal allowed reflection amplitude :math:`r`, must be in (0, 1)
        interval.
    degree : int
        order of the monominal absorbing boundary potential.
        Must be positive > 0.
    emin : float, optional
        lower energy cutoff. If set, only modes with energies above
        `emin` are considered. Defaut=None
    emax : float, optional
        upper energy cutoff. If set, only modes with energies below
        `emax` are considered. Defaut=None
    eps : float, optional
        minimal difference for length minimization step

    Returns
    -------
    num_cells : int
        number of cells for a finite lead taken into account
        `num_cells` is only meaningful if `absorbing_boundary` is `True`
    num_buffer_cells : int
        number of lead cells added to the start
        of the boundary conditions with no absorbing potential applied.
        `num_buffer_cells` is always meaningful,
        independent of `absorbing_boundary`
    strength : float
        The strength of the boundary conditions. Formally
        this is the area underneath the monomial curve.
        `strength` is only meaningful if `absorbing_boundary` is `True`
    absorbing_boundary : bool
        `True`: use absorbing boundary, optimal parameters were found
        `False`: use simple boundary, no optimal parameters were found

    Notes
    -------
    The routine returns the flag `absorbing_boundary` to tell the calling
    routine if (optimal) monomial parameters were found, in order
    to construct a monomial absorbing boundary condition.
    `absorbing_boundary = False` occurs if
    simple boundaries (only buffer cells) work better, or if the algorithm
    cannot find optimal parameters or if the algorithm fails.
    In all these casees simple boundaries serve as a save fallback,
    but maybe with a low performance.
    """
    if emin is not None:
        assert _common.is_type(emin, 'real_number')
    if emax is not None:
        assert _common.is_type(emax, 'real_number')
    if emax is not None and emin is not None:
        if emax < emin:
            raise ValueError('emax={} must be larger than emin={}.'
                             .format(emax, emin))

    # two reference points of the spectrum:
    # one with highest velocity (fast mode)
    # one with the highest curvature (slow mode)
    try:
        fast_mode = _fast_mode(spectrum, emin, emax)
        slow_mode = _slow_mode(spectrum, emin, emax)
    except Exception as error:
        # TODO: the case with emin/emax, where the highest velocity
        # resp. curvature do not follow analytically from the spectrum
        # is not yet handled properly.
        logger.warning('problem to analyse spectrum for boundary conditions: '
                       '{0}'.format(error))
        logger.info('switch to simple boundary conditions fallback')
        return 0, 0., 0, False

    _, vmax, _ = fast_mode

    assert vmax >= 0
    assert tmax >= 0
    assert eps >= 0

    len_max = vmax * tmax / 2
    length = len_max

    while True:
        length_new, strength, success = _new_length_estimate(fast_mode, slow_mode,
                                                             tmax, refl_max,
                                                             degree, length)
        logger.debug('length_new={}, strength={}, success={}'.
                     format(length_new, strength, success))
        if success and length_new <= length - eps:
            length = length_new
        else:
            break

    xsplit = _optimal_split(degree) if length < len_max else 1
    num_cells = ceil(length * (1 - xsplit))
    num_buffer_cells = ceil(length * xsplit)

    assert num_cells >= 0
    assert num_buffer_cells >= 0

    absorbing_boundary = length < len_max and num_cells > 0

    return num_cells, strength, num_buffer_cells, absorbing_boundary


@log_func
def automatic_boundary(leads, tmax, refl_max=1E-6, degree=6, emin=None,
                       emax=None, params=None):
    """
    Routine to find automatically a boundary condition such that the
    reflection amplitude :math:`r` for a lead stays below a given value.

    Parameters
    ----------
    leads : `kwant.system.InfiniteSystem` or `kwantspectrum.BandSketching`
        Kwant lead or `BandSketching` instance of the lead,
        for which the boundary condition is intended.
    tmax : int or float
        Maximal time for a tkwant simulation to run. Must be positive > 0.
    refl_max : float, optional
        maximal allowed reflection amplitude :math:`r`, must be in (0, 1)
        interval. Default=1E-7
    degree : int, optional
        order of the monominal absorbing boundary potential.
        Must be positive > 0. Default=4.
    emin : float, optional
        lower energy cutoff. If set, only modes with energies above
        `emin` are considered. Defaut=None
    emax : float, optional
        upper energy cutoff. If set, only modes with energies below
        `emax` are considered. Defaut=None
    params : dict, optional
        Extra arguments to pass to the Hamiltonian of ``leads``.
        Might only be provided if ``leads`` is an instance
        of `kwant.system.InfiniteSystem` (or a sequence thereof).
        If provided, ``params`` must include the time argument at initial time
        explicitly, if the lead Hamiltonian is explicitly time dependent.


    Returns
    -------
    boundaries : list of `MonomialAbsorbingBoundary` or `SimpleBoundary`
        List of boundary conditions. The length of ``boundaries`` is
        similar to the length of ``leads``.

    Notes
    -----
    The routine returns `MonomialAbsorbingBoundary` or `SimpleBoundary`
    conditions depending on which one is estimated to be computationally
    more efficient.

    """
    def get_boundary(i, lead):
        """Boundary condition for one lead"""
        if isinstance(lead, kwant.system.InfiniteSystem):
            logger.info('calculate lead spectrum for lead={}'.format(i))
            spectrum = kwantspectrum.spectrum(lead, params=params)
        else:  # assume that lead behaves like a spectrum
            spectrum = lead

        logger.info('estimate absorbing boundary parameters for lead={}'.format(i))
        num_cells, strength, num_buffer_cells, absorbing_boundary = \
            _monomial_parameter_estimate(spectrum, tmax, refl_max, degree,
                                         emin, emax)

        if absorbing_boundary:
            logger.info('use absorbing boundary for lead={}: num_cells = {}, '
                        'strength = {}, num_buffer_cells = {}'
                        .format(i, num_cells, strength, num_buffer_cells))
            return MonomialAbsorbingBoundary(num_cells, strength, degree,
                                             num_buffer_cells)
        logger.info('use simple boundary for lead={}'.format(i))
        return SimpleBoundary(tmax=tmax)

    logger.info('estimate optimal boundary conditions with parameters: '
                'tmax = {}, refl_max = {}, degree={}, emin={}, emax={}, '
                'params={}'.format(tmax, refl_max, degree, emin, emax, params))

    # type and consistency checks
    assert _common.is_type(tmax, 'real_number')
    assert _common.is_type(refl_max, 'real_number')
    assert _common.is_type(degree, 'real_number')
    if tmax <= 0:
        raise ValueError('Maximum time must be positive.')

    if not isinstance(leads, collections.abc.Iterable):
        leads = [leads]

    return [get_boundary(i, lead) for i, lead in enumerate(leads)]



# -*- coding: utf-8 -*-

# library-specific init
IF SPARSELIB == "rsb":
    cdef extern from "rsb.h":
        ctypedef signed int rsb_err_t;
        cdef struct rsb_initopts:
            pass
        cdef rsb_err_t rsb_lib_init(rsb_initopts *iop)
    # initialize librsb with no explicit options
    if rsb_lib_init(NULL) != 0:
        raise RuntimeError('Failed to initialize librsb')



# -*- coding: utf-8 -*-
# Copyright 2020 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Test module for `tkwant._logging`"""

import logging

from .. import _logging


def test_logging_level():
    assert _logging.level == logging.WARNING


def test_logging_handler():
    assert isinstance(_logging.handler, logging.Handler)
    assert isinstance(_logging.debug_handler, logging.Handler)
    assert isinstance(_logging.simple_handler, logging.Handler)


def test_logging_filter():
    assert _logging.filter is None


def test_logger(caplog):

    # make a logger and test that it is working if we use a predefined handler
    logger = _logging.make_logger(name=__name__)
    _logging.handler = _logging.simple_handler
    # no log events for info
    logger.info('this is a logger info')
    assert len(caplog.text) == 0
    # log event for warning (default level)
    logger.warning('this is a logger warning')
    assert "this is a logger warning" in caplog.text

    # make a logger and test that that we can change the log level
    logger = _logging.make_logger(name=__name__)
    _logging.handler = _logging.simple_handler
    _logging.level = logging.INFO
    # an event must be logged now at log-level info
    logger.info('this is a logger info')
    assert "this is a logger info" in caplog.text

    # we could do the same tests also for the handler and the filter
    # but the caplog fixture catches the logging events even when the handler
    # is the nullhandler or the messages are filtered.



# -*- coding: utf-8 -*-
# Copyright 2016 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.
"""tkwant, a Python package for time-resolved quantum transport."""

import os

from ._common import version as __version__  # pragma: no flakes
from ._common import TkwantDeprecationWarning
from . import _logging as logging

__all__ = ['TkwantDeprecationWarning', 'logging']

for module in ('system', 'leads', 'onebody', 'manybody', 'mpi'):
    exec('from . import {0}'.format(module))
    __all__.append(module)


# to make stuff available in tkwant namespace
# available = [
#    ('system', [])
# ]

# for module, names in available:
#    exec('from .{0} import {1}'.format(module, ', '.join(names)))
#    __all__.extend(names)
# del available, module, names  # remove cruft from namespace


del module  # remove cruft from namespace


def test(verbose=True):
    """Run tkwant's unit tests."""
    import pytest
    return pytest.main([os.path.dirname(os.path.abspath(__file__)), '-s']
                       + (['-v'] if verbose else []))



# -*- coding: utf-8 -*-
# Copyright 2016-2019 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Test module for `tkwant.integration`"""

import math
import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal
import pytest

from .. import integration


# (function, (a, b), integral)
functions = [
    (lambda x: x**2, (-1, 1), 2 / 3),
    (lambda x: 2 * np.exp(-x**2) / math.sqrt(math.pi), (0, 1), math.erf(1)),
    (lambda x: np.sin(x), (0, math.pi), 2),
]


def test_calc_abscissas_and_weights():

    # test the numerics
    quadrature_methods = ['gausslegendre', 'kronrod']
    nb_points = [10, 15, 20]

    def test_fixed(method, function, n):
        f, (a, b), exact_result = function
        x, w = integration.calc_abscissas_and_weights(a, b, n, method)
        if w.ndim == 1:  # for weight-vector (eg. gauss-legendre)
            result = np.sum(w * f(x))
            assert_almost_equal(result, exact_result)
            assert x.shape == (n,)
            assert w.shape == (n,)
        else:  # for weight-matrix (gauss-kronrod)
            results = np.sum(w * f(x), axis=1)
            [assert_almost_equal(result, exact_result) for result in results]
            assert x.shape == (2 * n + 1,)
            assert w.shape == (2, 2 * n + 1)

    for method in quadrature_methods:
        for func in functions:
            for n in nb_points:
                test_fixed(method, func, n)

    to_test = integration.calc_abscissas_and_weights

    a, b = (-2, 2)
    n = 20
    quadrature = 'gausslegendre'

    # test error raises
    # bounds swapped
    pytest.raises(ValueError, to_test, a=b, b=a, n=n, quadrature=quadrature)
    # invalid number of points
    pytest.raises(ValueError, to_test, a, b, n=0, quadrature=quadrature)
    pytest.raises(ValueError, to_test, a, b, n=-1, quadrature=quadrature)
    # nonexisting method
    pytest.raises(NotImplementedError, to_test, a, b, n, quadrature='bla')


def test_trapezoidal_rule():

    x, w = integration._trapezoidal_rule(n=2)
    assert_array_almost_equal(x, [-1, 1])
    assert_array_almost_equal(w, [1, 1])

    x, w = integration._trapezoidal_rule(n=3)
    assert_array_almost_equal(x, [-1, 0, 1])
    assert_array_almost_equal(w, [0.5, 1, 0.5])

    x, w = integration._trapezoidal_rule(n=4)
    assert_array_almost_equal(x, [-1, - 1 / 3, 1 / 3, 1])
    assert_array_almost_equal(w, (1 / 3) * np.array([1, 2, 2, 1]))

    with pytest.raises(ValueError) as exc:
        integration._trapezoidal_rule(1)
    assert str(exc.value) == 'n=1 must >= 2'



# -*- coding: utf-8 -*-
# cython: embedsignature=True
#
# Copyright 2016, 2017 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

from libc.string cimport memset
import numpy as np

from ..linalg import blas_sparse
from ..linalg cimport blas_sparse as sb
from ..linalg cimport blas as db
from .kernels cimport CKernel

cdef complex complex_one = 1, complex_minus_i = -1j, complex_zero = 0,


cdef class SparseBlas(CKernel):
    """Evaluate the RHS of the Schrödinger equation using sparse blas.

    Only available if tkwant has been built against a sparse BLAS library.

    Parameters
    ----------
    H0 : `scipy.sparse.base`
        Hamiltonian at ``t=0`` (including any boundary conditions).
    W : callable
        Time-dependent part of the Hamiltonian. Typically the object returned
        by `tkwant.system.extract_perturbation`.
    psi_st : `~numpy.ndarray`, optional
        The wavefunction of the initial eigenstate defined over the central
        system (if starting in an initial eigenstate).

    Attributes
    ----------
    H0 : int
        Handle for a matrix inside the sparse blas library. The Hamiltonian
        at ``t=0``.
    W : callable
        Time-dependent part of the Hamiltonian. Typically the object returned
        by `tkwant.system.extract_perturbation`.
    psi_st : array of complex or `None`
        The wavefunction of the initial eigenstate defined over the central
        system (if starting in an initial eigenstate).
    size : int
        The size of the time-independent part of the Hamiltonian. This
        also sets the size of the solution vector.
    """

    cdef sb.blas_sparse_matrix H0
    cdef object W
    cdef object args
    cdef complex[:] psi_st, temp
    cdef int center_size

    def __init__(self, H0, W, args, complex[:] psi_st=None):
        pass

    def __cinit__(self, H0, W, args, complex[:] psi_st=None):
        # `self.c_self` has the same lifetime as `self`; don't increment refcount
        self.c_self = <void*> self
        self.c_rhs = SparseBlas._rhs
        self.size = H0.shape[0]
        self.nevals = 0
        self.psi_st = psi_st
        self.W = W
        self.args = args
        if W is not None:
            self.center_size = self.W.size
        else:
            self.center_size = self.size

        H0 = H0.tocoo()

        # create temporary array storage
        self.temp = np.empty((self.center_size,), dtype=complex)

        # create sparse blas matrix handle from Scipy sparse matrix
        cdef int err = 0
        cdef complex[::1] data = H0.data
        cdef int[::1] cols = H0.col, rows = H0.row
        self.H0 = sb.BLAS_zuscr_begin(H0.shape[0], H0.shape[1])
        if self.H0 == sb.blas_invalid_handle:
            raise RuntimeError('Error initializing sparse matrix.')
        # this copies the data, so there is no need to incref H0
        err = sb.BLAS_zuscr_insert_entries(self.H0, H0.nnz,
                                           <const void*>&data[0],
                                           <const int*>&rows[0],
                                           <const int*>&cols[0])
        if err:
            raise RuntimeError('Error inserting entries into sparse matrix.')

        err = sb.BLAS_zuscr_end(self.H0)
        if err:
            raise RuntimeError('Error when finalizing sparse matrix.')

    def __dealloc__(self):
        if sb.BLAS_usds(self.H0):
            raise RuntimeError('Error destroying sparse matrix')

    def __init__(self, H0, W, args, complex[:] psi_st=None):
        # taken care of by __cinit__
        pass

    def set_args(self, args):
        self.args = args

    # TODO: make this more efficient -- fewer copies and temp storage
    # Actually evaluate the RHS. This is implemented as a cdef static method so
    # that the signature conforms to that expected by the C-level Solver API.
    @staticmethod
    cdef void _rhs(const void *_self, const complex *psi, complex *dpsidt,
                   double time) except *:
        cdef SparseBlas self = <SparseBlas>_self
        memset(<void*>dpsidt, 0, self.size * sizeof(complex))
        ### H0 @ psi -> dpsidt
        sb.BLAS_zusmv(sb.blas_no_trans, <const void*>&complex_one,
                      self.H0, <const void*>psi, 1, <void*>dpsidt, 1)
        ### W(t) @ temp + dpsidt -> dpsidt
        cdef complex *temp = psi
        if self.psi_st is not None:
            # we need to act on `temp = psi + psi_st`
            temp = &self.temp[0]
            # we only need construct vectors over the *central* system
            db.cblas_zcopy(self.center_size, <const void*>psi, 1,
                           temp, 1)
            db.cblas_zaxpy(self.center_size, <const void*>&complex_one,
                           <const void*>&self.psi_st[0], 1,
                           temp, 1)
        # finally act with the time-depedent part
        try:
            self.W.apply(time, <complex[:self.center_size]>temp,
                   out=<complex[:self.center_size]>dpsidt)
        except Exception as exept:
            if self.W is None:
                pass
            else:
                raise exept

        ### dpsidt = -1j * dpsidt
        db.cblas_zscal(self.size, <const void*>&complex_minus_i, <void*>dpsidt, 1)



ctypedef size_t CBLAS_INDEX

cdef extern from "cblas.h":
    ### Enumerated and derived types


    cdef enum CBLAS_ORDER:
        CblasRowMajor=101
        CblasColMajor=102

    cdef enum CBLAS_TRANSPOSE:
        CblasNoTrans=111
        CblasTrans=112
        CblasConjTrans=113

    cdef enum CBLAS_UPLO:
        CblasUpper=121
        CblasLower=122

    cdef enum CBLAS_DIAG:
        CblasNonUnit=131
        CblasUnit=132

    cdef enum CBLAS_SIDE:
        CblasLeft=141
        CblasRight=142

    # ===========================================================================
    # Prototypes for level 1 BLAS functions (complex are recast as routines)
    # ===========================================================================

    cdef float  cblas_sdsdot(const int N, const float alpha, const float *X,
                        const int incX, const float *Y, const int incY)
    cdef double cblas_dsdot(const int N, const float *X, const int incX, const float *Y,
                       const int incY)
    cdef float  cblas_sdot(const int N, const float  *X, const int incX,
                      const float  *Y, const int incY)
    cdef double cblas_ddot(const int N, const double *X, const int incX,
                      const double *Y, const int incY)

    ### Functions having prefixes Z and C only
    cdef void   cblas_cdotu_sub(const int N, const void *X, const int incX,
                           const void *Y, const int incY, void *dotu)
    cdef void   cblas_cdotc_sub(const int N, const void *X, const int incX,
                           const void *Y, const int incY, void *dotc)

    cdef void   cblas_zdotu_sub(const int N, const void *X, const int incX,
                           const void *Y, const int incY, void *dotu)
    cdef void   cblas_zdotc_sub(const int N, const void *X, const int incX,
                           const void *Y, const int incY, void *dotc)


    ### Functions having prefixes S D SC DZ
    cdef float  cblas_snrm2(const int N, const float *X, const int incX)
    cdef float  cblas_sasum(const int N, const float *X, const int incX)

    cdef double cblas_dnrm2(const int N, const double *X, const int incX)
    cdef double cblas_dasum(const int N, const double *X, const int incX)

    cdef float  cblas_scnrm2(const int N, const void *X, const int incX)
    cdef float  cblas_scasum(const int N, const void *X, const int incX)

    cdef double cblas_dznrm2(const int N, const void *X, const int incX)
    cdef double cblas_dzasum(const int N, const void *X, const int incX)


    ### Functions having standard 4 prefixes (S D C Z)
    cdef CBLAS_INDEX cblas_isamax(const int N, const float  *X, const int incX)
    cdef CBLAS_INDEX cblas_idamax(const int N, const double *X, const int incX)
    cdef CBLAS_INDEX cblas_icamax(const int N, const void   *X, const int incX)
    cdef CBLAS_INDEX cblas_izamax(const int N, const void   *X, const int incX)

    # ===========================================================================
    # Prototypes for level 1 BLAS routines
    # ===========================================================================

    ### Routines with standard 4 prefixes (s, d, c, z)
    cdef void cblas_sswap(const int N, float *X, const int incX,
                     float *Y, const int incY)
    cdef void cblas_scopy(const int N, const float *X, const int incX,
                     float *Y, const int incY)
    cdef void cblas_saxpy(const int N, const float alpha, const float *X,
                     const int incX, float *Y, const int incY)

    cdef void cblas_dswap(const int N, double *X, const int incX,
                     double *Y, const int incY)
    cdef void cblas_dcopy(const int N, const double *X, const int incX,
                     double *Y, const int incY)
    cdef void cblas_daxpy(const int N, const double alpha, const double *X,
                     const int incX, double *Y, const int incY)

    cdef void cblas_cswap(const int N, void *X, const int incX,
                     void *Y, const int incY)
    cdef void cblas_ccopy(const int N, const void *X, const int incX,
                     void *Y, const int incY)
    cdef void cblas_caxpy(const int N, const void *alpha, const void *X,
                     const int incX, void *Y, const int incY)

    cdef void cblas_zswap(const int N, void *X, const int incX,
                     void *Y, const int incY)
    cdef void cblas_zcopy(const int N, const void *X, const int incX,
                     void *Y, const int incY)
    cdef void cblas_zaxpy(const int N, const void *alpha, const void *X,
                     const int incX, void *Y, const int incY)


    # Routines with S and D prefix only
    cdef void cblas_srotg(float *a, float *b, float *c, float *s)
    cdef void cblas_srotmg(float *d1, float *d2, float *b1, const float b2, float *P)
    cdef void cblas_srot(const int N, float *X, const int incX,
                    float *Y, const int incY, const float c, const float s)
    cdef void cblas_srotm(const int N, float *X, const int incX,
                    float *Y, const int incY, const float *P)

    cdef void cblas_drotg(double *a, double *b, double *c, double *s)
    cdef void cblas_drotmg(double *d1, double *d2, double *b1, const double b2, double *P)
    cdef void cblas_drot(const int N, double *X, const int incX,
                    double *Y, const int incY, const double c, const double  s)
    cdef void cblas_drotm(const int N, double *X, const int incX,
                    double *Y, const int incY, const double *P)


    # Routines with S D C Z CS and ZD prefixes
    cdef void cblas_sscal(const int N, const float alpha, float *X, const int incX)
    cdef void cblas_dscal(const int N, const double alpha, double *X, const int incX)
    cdef void cblas_cscal(const int N, const void *alpha, void *X, const int incX)
    cdef void cblas_zscal(const int N, const void *alpha, void *X, const int incX)
    cdef void cblas_csscal(const int N, const float alpha, void *X, const int incX)
    cdef void cblas_zdscal(const int N, const double alpha, void *X, const int incX)


    # ===========================================================================
    # Prototypes for level 2 BLAS
    # ===========================================================================


    ### Routines with standard 4 prefixes (S, D, C, Z)
    cdef void cblas_sgemv(const  CBLAS_ORDER order,
                     const  CBLAS_TRANSPOSE TransA, const int M, const int N,
                     const float alpha, const float *A, const int lda,
                     const float *X, const int incX, const float beta,
                     float *Y, const int incY)
    cdef void cblas_sgbmv(const  CBLAS_ORDER order,
                     const  CBLAS_TRANSPOSE TransA, const int M, const int N,
                     const int KL, const int KU, const float alpha,
                     const float *A, const int lda, const float *X,
                     const int incX, const float beta, float *Y, const int incY)
    cdef void cblas_strmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const float *A, const int lda,
                     float *X, const int incX)
    cdef void cblas_stbmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const int K, const float *A, const int lda,
                     float *X, const int incX)
    cdef void cblas_stpmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const float *Ap, float *X, const int incX)
    cdef void cblas_strsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const float *A, const int lda, float *X,
                     const int incX)
    cdef void cblas_stbsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const int K, const float *A, const int lda,
                     float *X, const int incX)
    cdef void cblas_stpsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const float *Ap, float *X, const int incX)

    cdef void cblas_dgemv(const  CBLAS_ORDER order,
                     const  CBLAS_TRANSPOSE TransA, const int M, const int N,
                     const double alpha, const double *A, const int lda,
                     const double *X, const int incX, const double beta,
                     double *Y, const int incY)
    cdef void cblas_dgbmv(const  CBLAS_ORDER order,
                     const  CBLAS_TRANSPOSE TransA, const int M, const int N,
                     const int KL, const int KU, const double alpha,
                     const double *A, const int lda, const double *X,
                     const int incX, const double beta, double *Y, const int incY)
    cdef void cblas_dtrmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const double *A, const int lda,
                     double *X, const int incX)
    cdef void cblas_dtbmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const int K, const double *A, const int lda,
                     double *X, const int incX)
    cdef void cblas_dtpmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const double *Ap, double *X, const int incX)
    cdef void cblas_dtrsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const double *A, const int lda, double *X,
                     const int incX)
    cdef void cblas_dtbsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const int K, const double *A, const int lda,
                     double *X, const int incX)
    cdef void cblas_dtpsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const double *Ap, double *X, const int incX)

    cdef void cblas_cgemv(const  CBLAS_ORDER order,
                     const  CBLAS_TRANSPOSE TransA, const int M, const int N,
                     const void *alpha, const void *A, const int lda,
                     const void *X, const int incX, const void *beta,
                     void *Y, const int incY)
    cdef void cblas_cgbmv(const  CBLAS_ORDER order,
                     const  CBLAS_TRANSPOSE TransA, const int M, const int N,
                     const int KL, const int KU, const void *alpha,
                     const void *A, const int lda, const void *X,
                     const int incX, const void *beta, void *Y, const int incY)
    cdef void cblas_ctrmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const void *A, const int lda,
                     void *X, const int incX)
    cdef void cblas_ctbmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const int K, const void *A, const int lda,
                     void *X, const int incX)
    cdef void cblas_ctpmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const void *Ap, void *X, const int incX)
    cdef void cblas_ctrsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const void *A, const int lda, void *X,
                     const int incX)
    cdef void cblas_ctbsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const int K, const void *A, const int lda,
                     void *X, const int incX)
    cdef void cblas_ctpsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const void *Ap, void *X, const int incX)

    cdef void cblas_zgemv(const  CBLAS_ORDER order,
                     const  CBLAS_TRANSPOSE TransA, const int M, const int N,
                     const void *alpha, const void *A, const int lda,
                     const void *X, const int incX, const void *beta,
                     void *Y, const int incY)
    cdef void cblas_zgbmv(const  CBLAS_ORDER order,
                     const  CBLAS_TRANSPOSE TransA, const int M, const int N,
                     const int KL, const int KU, const void *alpha,
                     const void *A, const int lda, const void *X,
                     const int incX, const void *beta, void *Y, const int incY)
    cdef void cblas_ztrmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const void *A, const int lda,
                     void *X, const int incX)
    cdef void cblas_ztbmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const int K, const void *A, const int lda,
                     void *X, const int incX)
    cdef void cblas_ztpmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const void *Ap, void *X, const int incX)
    cdef void cblas_ztrsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const void *A, const int lda, void *X,
                     const int incX)
    cdef void cblas_ztbsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const int K, const void *A, const int lda,
                     void *X, const int incX)
    cdef void cblas_ztpsv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE TransA, const  CBLAS_DIAG Diag,
                     const int N, const void *Ap, void *X, const int incX)


    ### Routines with S and D prefixes only
    cdef void cblas_ssymv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const float alpha, const float *A,
                     const int lda, const float *X, const int incX,
                     const float beta, float *Y, const int incY)
    cdef void cblas_ssbmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const int K, const float alpha, const float *A,
                     const int lda, const float *X, const int incX,
                     const float beta, float *Y, const int incY)
    cdef void cblas_sspmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const float alpha, const float *Ap,
                     const float *X, const int incX,
                     const float beta, float *Y, const int incY)
    cdef void cblas_sger(const  CBLAS_ORDER order, const int M, const int N,
                    const float alpha, const float *X, const int incX,
                    const float *Y, const int incY, float *A, const int lda)
    cdef void cblas_ssyr(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const float alpha, const float *X,
                    const int incX, float *A, const int lda)
    cdef void cblas_sspr(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const float alpha, const float *X,
                    const int incX, float *Ap)
    cdef void cblas_ssyr2(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const float alpha, const float *X,
                    const int incX, const float *Y, const int incY, float *A,
                    const int lda)
    cdef void cblas_sspr2(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const float alpha, const float *X,
                    const int incX, const float *Y, const int incY, float *A)

    cdef void cblas_dsymv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const double alpha, const double *A,
                     const int lda, const double *X, const int incX,
                     const double beta, double *Y, const int incY)
    cdef void cblas_dsbmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const int K, const double alpha, const double *A,
                     const int lda, const double *X, const int incX,
                     const double beta, double *Y, const int incY)
    cdef void cblas_dspmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const double alpha, const double *Ap,
                     const double *X, const int incX,
                     const double beta, double *Y, const int incY)
    cdef void cblas_dger(const  CBLAS_ORDER order, const int M, const int N,
                    const double alpha, const double *X, const int incX,
                    const double *Y, const int incY, double *A, const int lda)
    cdef void cblas_dsyr(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const double alpha, const double *X,
                    const int incX, double *A, const int lda)
    cdef void cblas_dspr(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const double alpha, const double *X,
                    const int incX, double *Ap)
    cdef void cblas_dsyr2(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const double alpha, const double *X,
                    const int incX, const double *Y, const int incY, double *A,
                    const int lda)
    cdef void cblas_dspr2(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const double alpha, const double *X,
                    const int incX, const double *Y, const int incY, double *A)


    ### Routines with C and Z prefixes only
    cdef void cblas_chemv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const void *alpha, const void *A,
                     const int lda, const void *X, const int incX,
                     const void *beta, void *Y, const int incY)
    cdef void cblas_chbmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const int K, const void *alpha, const void *A,
                     const int lda, const void *X, const int incX,
                     const void *beta, void *Y, const int incY)
    cdef void cblas_chpmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const void *alpha, const void *Ap,
                     const void *X, const int incX,
                     const void *beta, void *Y, const int incY)
    cdef void cblas_cgeru(const  CBLAS_ORDER order, const int M, const int N,
                     const void *alpha, const void *X, const int incX,
                     const void *Y, const int incY, void *A, const int lda)
    cdef void cblas_cgerc(const  CBLAS_ORDER order, const int M, const int N,
                     const void *alpha, const void *X, const int incX,
                     const void *Y, const int incY, void *A, const int lda)
    cdef void cblas_cher(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const float alpha, const void *X, const int incX,
                    void *A, const int lda)
    cdef void cblas_chpr(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const float alpha, const void *X,
                    const int incX, void *A)
    cdef void cblas_cher2(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo, const int N,
                    const void *alpha, const void *X, const int incX,
                    const void *Y, const int incY, void *A, const int lda)
    cdef void cblas_chpr2(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo, const int N,
                    const void *alpha, const void *X, const int incX,
                    const void *Y, const int incY, void *Ap)

    cdef void cblas_zhemv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const void *alpha, const void *A,
                     const int lda, const void *X, const int incX,
                     const void *beta, void *Y, const int incY)
    cdef void cblas_zhbmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const int K, const void *alpha, const void *A,
                     const int lda, const void *X, const int incX,
                     const void *beta, void *Y, const int incY)
    cdef void cblas_zhpmv(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                     const int N, const void *alpha, const void *Ap,
                     const void *X, const int incX,
                     const void *beta, void *Y, const int incY)
    cdef void cblas_zgeru(const  CBLAS_ORDER order, const int M, const int N,
                     const void *alpha, const void *X, const int incX,
                     const void *Y, const int incY, void *A, const int lda)
    cdef void cblas_zgerc(const  CBLAS_ORDER order, const int M, const int N,
                     const void *alpha, const void *X, const int incX,
                     const void *Y, const int incY, void *A, const int lda)
    cdef void cblas_zher(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const double alpha, const void *X, const int incX,
                    void *A, const int lda)
    cdef void cblas_zhpr(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo,
                    const int N, const double alpha, const void *X,
                    const int incX, void *A)
    cdef void cblas_zher2(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo, const int N,
                    const void *alpha, const void *X, const int incX,
                    const void *Y, const int incY, void *A, const int lda)
    cdef void cblas_zhpr2(const  CBLAS_ORDER order, const  CBLAS_UPLO Uplo, const int N,
                    const void *alpha, const void *X, const int incX,
                    const void *Y, const int incY, void *Ap)


    # ===========================================================================
    # Prototypes for level 3 BLAS
    # ===========================================================================


    ### Routines with standard 4 prefixes (S, D, C, Z)
    cdef void cblas_sgemm(const  CBLAS_ORDER Order, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_TRANSPOSE TransB, const int M, const int N,
                     const int K, const float alpha, const float *A,
                     const int lda, const float *B, const int ldb,
                     const float beta, float *C, const int ldc)
    cdef void cblas_ssymm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const int M, const int N,
                     const float alpha, const float *A, const int lda,
                     const float *B, const int ldb, const float beta,
                     float *C, const int ldc)
    cdef void cblas_ssyrk(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                     const float alpha, const float *A, const int lda,
                     const float beta, float *C, const int ldc)
    cdef void cblas_ssyr2k(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                      const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                      const float alpha, const float *A, const int lda,
                      const float *B, const int ldb, const float beta,
                      float *C, const int ldc)
    cdef void cblas_strmm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_DIAG Diag, const int M, const int N,
                     const float alpha, const float *A, const int lda,
                     float *B, const int ldb)
    cdef void cblas_strsm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_DIAG Diag, const int M, const int N,
                     const float alpha, const float *A, const int lda,
                     float *B, const int ldb)

    cdef void cblas_dgemm(const  CBLAS_ORDER Order, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_TRANSPOSE TransB, const int M, const int N,
                     const int K, const double alpha, const double *A,
                     const int lda, const double *B, const int ldb,
                     const double beta, double *C, const int ldc)
    cdef void cblas_dsymm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const int M, const int N,
                     const double alpha, const double *A, const int lda,
                     const double *B, const int ldb, const double beta,
                     double *C, const int ldc)
    cdef void cblas_dsyrk(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                     const double alpha, const double *A, const int lda,
                     const double beta, double *C, const int ldc)
    cdef void cblas_dsyr2k(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                      const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                      const double alpha, const double *A, const int lda,
                      const double *B, const int ldb, const double beta,
                      double *C, const int ldc)
    cdef void cblas_dtrmm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_DIAG Diag, const int M, const int N,
                     const double alpha, const double *A, const int lda,
                     double *B, const int ldb)
    cdef void cblas_dtrsm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_DIAG Diag, const int M, const int N,
                     const double alpha, const double *A, const int lda,
                     double *B, const int ldb)

    cdef void cblas_cgemm(const  CBLAS_ORDER Order, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_TRANSPOSE TransB, const int M, const int N,
                     const int K, const void *alpha, const void *A,
                     const int lda, const void *B, const int ldb,
                     const void *beta, void *C, const int ldc)
    cdef void cblas_csymm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const int M, const int N,
                     const void *alpha, const void *A, const int lda,
                     const void *B, const int ldb, const void *beta,
                     void *C, const int ldc)
    cdef void cblas_csyrk(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                     const void *alpha, const void *A, const int lda,
                     const void *beta, void *C, const int ldc)
    cdef void cblas_csyr2k(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                      const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                      const void *alpha, const void *A, const int lda,
                      const void *B, const int ldb, const void *beta,
                      void *C, const int ldc)
    cdef void cblas_ctrmm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_DIAG Diag, const int M, const int N,
                     const void *alpha, const void *A, const int lda,
                     void *B, const int ldb)
    cdef void cblas_ctrsm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_DIAG Diag, const int M, const int N,
                     const void *alpha, const void *A, const int lda,
                     void *B, const int ldb)

    cdef void cblas_zgemm(const  CBLAS_ORDER Order, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_TRANSPOSE TransB, const int M, const int N,
                     const int K, const void *alpha, const void *A,
                     const int lda, const void *B, const int ldb,
                     const void *beta, void *C, const int ldc)
    cdef void cblas_zsymm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const int M, const int N,
                     const void *alpha, const void *A, const int lda,
                     const void *B, const int ldb, const void *beta,
                     void *C, const int ldc)
    cdef void cblas_zsyrk(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                     const void *alpha, const void *A, const int lda,
                     const void *beta, void *C, const int ldc)
    cdef void cblas_zsyr2k(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                      const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                      const void *alpha, const void *A, const int lda,
                      const void *B, const int ldb, const void *beta,
                      void *C, const int ldc)
    cdef void cblas_ztrmm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_DIAG Diag, const int M, const int N,
                     const void *alpha, const void *A, const int lda,
                     void *B, const int ldb)
    cdef void cblas_ztrsm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const  CBLAS_TRANSPOSE TransA,
                     const  CBLAS_DIAG Diag, const int M, const int N,
                     const void *alpha, const void *A, const int lda,
                     void *B, const int ldb)


    ### Routines with prefixes C and Z only
    cdef void cblas_chemm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const int M, const int N,
                     const void *alpha, const void *A, const int lda,
                     const void *B, const int ldb, const void *beta,
                     void *C, const int ldc)
    cdef void cblas_cherk(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                     const float alpha, const void *A, const int lda,
                     const float beta, void *C, const int ldc)
    cdef void cblas_cher2k(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                      const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                      const void *alpha, const void *A, const int lda,
                      const void *B, const int ldb, const float beta,
                      void *C, const int ldc)

    cdef void cblas_zhemm(const  CBLAS_ORDER Order, const  CBLAS_SIDE Side,
                     const  CBLAS_UPLO Uplo, const int M, const int N,
                     const void *alpha, const void *A, const int lda,
                     const void *B, const int ldb, const void *beta,
                     void *C, const int ldc)
    cdef void cblas_zherk(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                     const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                     const double alpha, const void *A, const int lda,
                     const double beta, void *C, const int ldc)
    cdef void cblas_zher2k(const  CBLAS_ORDER Order, const  CBLAS_UPLO Uplo,
                      const  CBLAS_TRANSPOSE Trans, const int N, const int K,
                      const void *alpha, const void *A, const int lda,
                      const void *B, const int ldb, const double beta,
                      void *C, const int ldc)

    cdef void cblas_xerbla(int p, const char *rout, const char *form, ...)



# -*- coding: utf-8 -*-
# Copyright 2016 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Test module for `tkwant.system`"""

import numpy as np
import kwant
import pytest

from .. import system, leads
from .common import (make_chain, make_simple_lead, make_complex_lead,
                     make_system_with_leads, check_boundary_hamiltonian)


def test_orbital_slices():

    def check(fsyst):
        slices = (system.orb_range(fsyst, i) for i in range(len(fsyst.sites)))
        stop = 0
        for site, (start_orb, stop_orb) in zip(fsyst.sites, slices):
            assert start_orb == stop
            stop += site.family.norbs
            assert stop_orb == stop
            assert site.family.norbs == (stop_orb - start_orb)

    # uniform families
    lat, syst = make_chain(3, norbs=1)
    check(syst.finalized())
    lat, syst = make_chain(3, norbs=2)
    check(syst.finalized())
    # mixed families
    lat = kwant.lattice.chain(name='a', norbs=1)
    lat2 = kwant.lattice.chain(name='b', norbs=2)
    syst = kwant.Builder()
    syst[map(lat, range(3))] = 2
    syst[map(lat2, range(3))] = 2 * np.eye(2)
    check(syst.finalized())


@pytest.mark.parametrize('lead_maker', [make_simple_lead, make_complex_lead])
def test_hamiltonian(lead_maker):
    ncells = 10
    strength = 10
    degree = 6
    # construct
    syst = make_system_with_leads(kwant.lattice.square(norbs=1),
                                  lead_maker)
    fsyst = syst.finalized()
    boundaries = [leads.SimpleBoundary(10),
                  leads.MonomialAbsorbingBoundary(ncells, strength, degree)]
    Hext = system.hamiltonian_with_boundaries(fsyst, boundaries, params={})

    # test shapes
    norbs_central = len(fsyst.sites)
    norbs_leads = [lead.cell_size * ncells for lead in fsyst.leads]
    assert Hext.hamiltonian.shape[0] == Hext.hamiltonian.shape[1]
    assert Hext.hamiltonian.shape[0] == norbs_central + sum(norbs_leads)

    # test absorbing regions and coupling

    def absorbing(i):
        n = degree
        return -1j * np.eye(norbs) * ((n + 1) * strength *
                                      i**n / ncells**(n + 1))

    loop = zip(fsyst.leads, fsyst.lead_interfaces, Hext.boundary_slices,
               boundaries)
    for lead, lead_interface, slc, bdy in loop:
        norbs = lead.cell_size
        norbs_iface = lead.graph.num_nodes - lead.cell_size
        V = lead.inter_cell_hopping()
        V_dag = V.conjugate().transpose()
        if isinstance(bdy, leads.MonomialAbsorbingBoundary):
            def onsite(i):
                return lead.cell_hamiltonian() + absorbing(i)
        else:
            def onsite(i):
                return lead.cell_hamiltonian()
        check_boundary_hamiltonian(Hext.hamiltonian[slc, slc],
                                   norbs, norbs_iface, onsite, lambda i: V)

        # test coupling -- uses fact that norbs=1
        for i, iface_site in enumerate(lead_interface):
            bdy_iface = slice(slc.start, slc.start + norbs)
            iface_slice = slice(iface_site, iface_site + 1)
            assert np.allclose(Hext.hamiltonian[bdy_iface, iface_slice].todense(),
                               V[:, i:(i + 1)])
            assert np.allclose(Hext.hamiltonian[iface_slice, bdy_iface].todense(),
                               V_dag[i:(i + 1), :])

    # test is_valid and should_continue
    psi_test = np.empty(Hext.hamiltonian.shape[0], dtype=complex)
    assert Hext.solution_is_valid(psi_test)  # should always return True
    vmax = np.linalg.norm(fsyst.leads[0].inter_cell_hopping(), ord=2)
    tmax = ncells / (2 * vmax)
    eps = 1E-8
    assert Hext.time_is_valid(tmax - eps)
    assert not Hext.time_is_valid(tmax + eps)


@pytest.mark.parametrize('lead_maker', [make_simple_lead])
def test_hamiltonian_with_boundaries_exceptions(lead_maker):

    # construct
    syst = make_system_with_leads(kwant.lattice.square(norbs=1), lead_maker)
    fsyst = syst.finalized()

    # too less boundary conditions
    boundaries = [leads.SimpleBoundary(10)]
    with pytest.raises(ValueError) as exc:
        system.hamiltonian_with_boundaries(fsyst, boundaries, params={})
    assert 'Number of leads= 2 does not match the number of boundaries provided= 1' in str(exc.value)

    # too many boundary conditions
    boundaries = [leads.SimpleBoundary(10)] * 3
    with pytest.raises(ValueError) as exc:
        system.hamiltonian_with_boundaries(fsyst, boundaries, params={})
    assert 'Number of leads= 2 does not match the number of boundaries provided= 3' in str(exc.value)


def test_is_time_dependent_function():
    def time_dependent(zeit):
        pass
    not_a_function = 'time'
    assert system.is_time_dependent_function(time_dependent, 'zeit')
    assert system.is_time_dependent_function(time_dependent, 'time') is False
    assert system.is_time_dependent_function(not_a_function, 'time') is False


def test_add_time_to_params():
    params_example = {'some_key': 2}
    time_name = 'my_time'
    time_value = 42

    # check against reference result
    tparams = system.add_time_to_params(params_example, time_name=time_name,
                                        time=time_value)
    assert isinstance(tparams, dict)
    assert tparams == {'some_key': 2, 'my_time': 42}

    # check for possible inputs
    for params in [None, {}, params_example]:
        tparams = system.add_time_to_params(params, time_name=time_name,
                                            time=time_value)
        assert isinstance(tparams, dict)
        assert time_name in tparams
        assert tparams[time_name] == time_value

        # check that having a dict with already the time input gives an error
        with pytest.raises(KeyError) as exc:
            system.add_time_to_params(tparams, time_name=time_name,
                                      time=time_value)
        error_string = str(exc.value).lstrip('\"').rstrip('\"')  # remove etra quotes
        assert error_string == "'params' must not contain my_time"

    # check that the (mutable) input dict is really copied and no reference created
    params_copy = params_example.copy()
    tparams = system.add_time_to_params(params_copy, time_name=time_name,
                                        time=time_value)
    tparams = {}
    assert params_copy == params_example



# -*- coding: utf-8 -*-
# Copyright 2016-2019 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.
"""Tools for solving the one-body time-dependent Schrödinger equation."""

__all__ = []

for module in ('solvers', 'kernels', 'onebody'):
    exec('from . import {0}'.format(module))
    __all__.append(module)

available = [
    ('onebody', onebody.__all__)
]

for module, names in available:
    exec('from .{0} import {1}'.format(module, ', '.join(names)))
    __all__.extend(names)

del available, module, names  # remove cruft from namespace



# -*- coding: utf-8 -*-
# Copyright 2020 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.
"""Logging routines."""

import logging
import functools
import sys

__all__ = ['level', 'handler', 'filter', 'debug_handler', 'simple_handler',
           'make_logger', 'log_func']


def _make_handler(format=None):
    """Return a stream handler from a given format"""
    handler = logging.StreamHandler(sys.stdout)
    if format is not None:
        handler.setFormatter(format)
    return handler


def _make_logger(name, level, handler, filter=None):
    """Return a logger"""
    _logger = logging.getLogger(name)
    _logger.setLevel(level)
    _logger.addHandler(handler)
    if filter is not None:
        _logger.addFilter(filter)
    return _logger


def make_logger(name):
    """Return a logger with additional information

    Parameters
    ----------
    name : str
        The name of the logger
    info : callable
        A function returning additional information as a dict.
        Dict attributes appear as LogRecord attributes in the logger
        and can be included in the logger output with an appropriate format.

    Returns
    -------
    DelayedLogger : `logging.Logger`
        Logger instance with additional information and delayed creation.

    Notes
    -----
    This function uses the module attributes: level, filter, handler
    """
    class CustomLogger(logging.Logger):
        """Customize the logger class with additional information."""
        def _log(self, level, msg, args, exc_info=None, extra=None):
            if extra is None:
                extra = info()
            super(CustomLogger, self)._log(level, msg, args, exc_info, extra)

    class DelayedLogger():
        """An instance of this class works like the logger in python standard lib.

        This class delays the creation of the actual logger to the moment,
        when a first logging event is triggered.
        This allows to overwrite default level, handlers and filters,
        stored in this class, at runtime.
        """
        # the logger is a singleton, so we store it as class variable
        _logger = None

        def __getattribute__(self, key):
            # the first attribute access will initializes the logger
            try:
                return getattr(DelayedLogger._logger, key)
            except Exception as e:
                if DelayedLogger._logger is None:
                    logging.setLoggerClass(CustomLogger)  # add information to logger
                    DelayedLogger._logger = _make_logger(name, level, handler, filter)
                return self.__getattribute__(key)
                raise e

    from .mpi import get_rank  # importing at module level gives circular import
    info = get_rank
    assert callable(info)
    assert isinstance(name, str)
    return DelayedLogger()


def log_func(logger, funcname=''):
    r"""Return a wrapper to log function calls

    Use this wrapper as:

    @log_func(my_logger)
    def my_function():
        pass

    Parameters
    ----------
    logger : `logging.Logger`
            Logger instance
    funcname : str, optional
        Name of the function that is decorated that will apper in the log.

    Returns
    -------
    decorate : callable
        A wrapper to log function call and return.
    """
    def decorate(func):
        logname = funcname if funcname else func.__name__
        startmsg = 'start {}'.format(logname)
        endmsg = '{} finished'.format(logname)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            logger.debug(startmsg)
            result = func(*args, **kwargs)
            logger.debug(endmsg)
            return result
        return wrapper
    return decorate


# two predefined handlers for tkwant

"""logging handler with format: level:module-name:line-number:MPI-rank: message"""
_format = logging.Formatter('%(levelname)s:%(name)s:%(lineno)d:rank=%(rank)s: %(message)s')
debug_handler = _make_handler(_format)

"""logging handler with format: message"""
simple_handler = _make_handler()

# tkwant's default logging settings
level = logging.WARNING
handler = logging.NullHandler()  # suppress logging
filter = None
